unit rDBComponents_UDB;

{<
  rDBComponents_UDB (UDB= Universal DataBase) is set of enhanced DB components
  which can be used instead of standard db components and provide
  the same properties and functions for different types of databases.
  It allows to convert application very easilly from one database type to another.

  Now it supports following type of database connections and queries:
  ADO = dbGo (used for MSAccess, MSSQLServer nad other db via ODBC or JET)
  ZEOS (used for MySQL, Firebird and other supported databases)
  DBX = dbExpress (used for MySQL, Interbase, Firebird and other supported databases)
  IB = Interbase (used for Interbase or Firebird)
  FIB = FIBBlus  (used for Interbase or Firebird)
  FireDac

  Usage of particular DB type is set by compiler conditional defines:
  DEFINE USEADO  for ADO
  DEFINE USEZEOS for Zeos
  DEFINE USEDBX  for DBX
  DEFINE USEIB   for Interbase
  DEFINE USEFIB  for FIBBlus
  DEFINE USEFD  for FireDac


  Unit contains new data connection components:
  TConnectionUDB - descendant of TADOConnection/TZConnection/TSQLConnection/TIBDataBase/TpFIBDatabase/TFDConnection
  TQueryUDB - descendant of TADOQuery/TZQuery/TSimpleDataSet/TIBQuery/TpFIBDataSet/TFDQuery

  TQueryUDB provides many new functions and properties usefull for any
  db application for centralization of all db tasks:
  - new props (TableName, IdFieldName, FieldTagToDuplic, FieldTagToReadOnly...)
  - new function (DuplicateRecord, SetFieldsReadOnly - according of fields tag value)
  - new events for centralized setting of fields captions and properties in application
  - many new function for setting of SQL commands in runtime


  Unit contain set of visible DB components.
  All of them contain internal TQueryUDB dataset with definable SQL command
  for loading of list values and events for setting of SQL in run time,
  so no other auxilary dataset is needed on the form.
  It can show bounded buttons used for enhanced selection of value
  or showing details of selected record dependently on application.

  TEditUDB - TEdit descendant loads and display text from any SQL command
  TComboBoxUDB - TrComboBoxEx descendant load list of items from any SQL command
  TrDBComboBoxUDB - TrDBComboBox descendant load list of items from any SQL command
  TrDBLookupComboBoxUDB - TrDBLookupComboBox descendant load list of items from any SQL command
  TListBoxUDB - TListBox descendant load list of items from any SQL command
  TrDBListBoxUDB - TrDBListBox descendant load list of items from any SQL command
  TrDBLookupListBoxUDB - TrDBLookupListBox descendant load list of items from any SQL command

  TCheckListBoxUDB - TCheckListBox descendant with possibility to load list of items from any SQL command
  TrDBCheckListBoxUDB - TrDBCheckListBox descendant with possibility to load list of items and values
  from any SQL command, provide all functions for N:M relationship

  TrDBRecViewUDB - TrDBRecView descendant with possibility to load/show/edit record from any SQL command
  TrDBGrid_MSUDB - TrDBGrid_MS descendant with possibility to load/show/edit records from any SQL command

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2011)
  @lastmod(Nov 2020)
}

interface

// define UDB type
{$DEFINE UDB_FIBComp}
// define internal directives according to project directive
{$IFDEF UDB_UDBComp}
{$IF Defined(USEADO)}{$DEFINE UDB_ADO}{$IFEND}
{$IF Defined(USEDBX)}{$DEFINE UDB_DBX}{$IFEND}
{$IF Defined(USEIB)}{$DEFINE UDB_IB}{$IFEND}
{$IF Defined(USEFIB)}{$DEFINE UDB_FIB}{$IFEND}
{$IF Defined(USEFD)}{$DEFINE UDB_FD}{$IFEND}
{$IF Defined(USEZeos)}{$DEFINE UDB_Zeos}{$IFEND}
{$ENDIF}
{$IFDEF UDB_ADOComp}{$DEFINE UDB_ADO}{$ENDIF}
{$IFDEF UDB_DBXComp}{$DEFINE UDB_DBX}{$ENDIF}
{$IFDEF UDB_IBComp}{$DEFINE UDB_IB}{$ENDIF}
{$IFDEF UDB_FIBComp}{$DEFINE UDB_FIB}{$ENDIF}
{$IFDEF UDB_FDComp}{$DEFINE UDB_FD}{$ENDIF}
{$IFDEF UDB_ZeosComp}{$DEFINE UDB_Zeos}{$ENDIF}
// define other supplement directives
{$DEFINE DoNotUseGlyphInDesign}
{$IF Defined(UDB_ADO) OR Defined(UDB_DBX) OR Defined(UDB_IB) OR Defined(UDB_FIB) OR Defined(UDB_FD)}
{$DEFINE UDB_AddConReadOnly}
{$DEFINE UDB_AddConHostName}
{$IFEND}
{$IF Defined(UDB_ADO) OR Defined(UDB_IB) OR Defined(UDB_FIB) OR Defined(UDB_FD)}
{$DEFINE UDB_AddQueryReadOnly}
{$IFEND}
{$IF Defined(UDB_DBX) OR Defined(UDB_FIB)}
{$DEFINE UDB_AddQuerySQL}
{$IFEND}
{$IF Defined(UDB_IB) OR Defined(UDB_FIB)}
{$DEFINE UDB_AddQueryConnection}
{$IFEND}

uses
  Windows, Messages, SysUtils, Graphics, Classes, Controls, Menus, Dialogs,
  StdCtrls, ExtCtrls, Buttons, Forms, DB, Variants, CheckLst,
{$IFDEF UDB_ADO} Adodb, {$ENDIF}
{$IFDEF UDB_Zeos} ZConnection, ZAbstractRODataset, ZAbstractDataset, ZDataSet, {$ENDIF}
{$IFDEF UDB_DBX} SqlExpr, DBClient, SimpleDS, {$ENDIF}
{$IFDEF UDB_IB} IBDatabase, IBCustomDataSet, IBQuery, {$ENDIF}
{$IFDEF UDB_FIB} FIBDatabase, pFIBDatabase, FIBDataSet, pFIBDataSet, {$ENDIF}
{$IFDEF UDB_FD} FireDAC.Comp.DataSet, FireDAC.Comp.Client, {$ENDIF}
  rDBComponents, rImprovedComps, rDBRecView, rDBGrid, rDBGrid_MS;

type
  TCustomConnectionUDB = class;
  TCustomQueryUDB = class;

  { list of dataset actions stored as last executed action }
  TDatasetLastEditAction = (leNone, leEdit, leInsert, leDuplic, leCancel, leDelete);
  { list of dataset event types }
  TDatasetEventType = (deNone, deBeforeOpen, deAfterOpen, deBeforeClose, deAfterClose, deBeforePost, deAfterPost, deBeforeScroll, deAfterScroll,
    deBeforeCancel, deAfterCancel, deBeforeInsert, deAfterInsert, deBeforeEdit, deAfterEdit, deBeforeDelete, deAfterDelete, deBeforeApplyUpdates,
    deAfterApplyUpdates, deAfterPostOrApplyUpdates);
  { set of dataset event types }
  TDatasetEventTypes = set of TDatasetEventType;

  { event type used by UDBComponents }
  TDataSetEvent = procedure(DataSet: TDataSet; EventType: TDatasetEventType = deNone) of object;
  { event type used by TCustomQueryUDB to get SQL command }
  TGetSQLEvent = procedure(Sender: TObject; SQL: TStrings) of object;
  { event type used by UDBComponents after click on bound button }
  TClickComboButton = procedure(Sender: TObject; Con: TCustomConnectionUDB; TableName: string; DataField, ListField: TField) of object;

  { TCustomConnectionUDB - descendant of TADOConnection/TZConnection/TSQLConnection/TIBDataBase/TpFIBDatabase
    which provides same functions for all connection types}
  TCustomConnectionUDB = class(
{$IFDEF UDB_ADO} TADOConnection {$ENDIF}
{$IFDEF UDB_Zeos} TZConnection {$ENDIF}
{$IFDEF UDB_DBX} TSQLConnection {$ENDIF}
{$IFDEF UDB_IB} TIBDataBase {$ENDIF}
{$IFDEF UDB_FIB} TpFIBDatabase {$ENDIF}
{$IFDEF UDB_FD} TFDConnection {$ENDIF})
  private
{$IFDEF UDB_ADO}
    FConnectionStringOrig: string;
    FConnectionStringDesign: string;
{$ENDIF}
{$IFDEF UDB_AddConReadOnly}
    FReadOnly: boolean;
    procedure SetReadOnly(Value: boolean);
{$ENDIF}
{$IFDEF UDB_AddConHostName}
    function GetHostName: string;
{$ENDIF}
  protected
{$IFDEF UDB_ADO}
    procedure DoConnect; override;
{$ENDIF}
  public
    { return database name }
    function GetDatabaseName: string;
    { scan all registered dataset and call CallBackProc for each one }
    procedure CallProcForRegisteredDataset(CallBackProc: TNotifyEvent);
{$IFDEF UDB_Zeos}
    { open connection }
    procedure Open;
    { close connection }
    procedure Close;
{$ENDIF}
  published
{$IFDEF UDB_ADO}
    { copy of connection string in design time }
    property ConnectionStringDesign: string read FConnectionStringDesign write FConnectionStringDesign;
{$ENDIF}
{$IFDEF UDB_AddConReadOnly}
    property ReadOnly: boolean read FReadOnly write SetReadOnly default false;
{$ENDIF}
{$IFDEF UDB_AddConHostName}
    property HostName: string read GetHostName stored false;
{$ENDIF}
{$IFDEF UDB_Zeos}
    { list of registered datasets }
    property RegisteredDatasets: TList read FDataSets;
{$ENDIF}
  end;

  { TCustomQueryUDB - descendant of TADOQuery/TZQuery/TSimpleDataSet/TIBQuery/TpFIBDataSet
    which provides same functions for all dataset types}
  TCustomQueryUDB = class(
{$IFDEF UDB_ADO} TADOQuery {$ENDIF}
{$IFDEF UDB_Zeos} TZQuery {$ENDIF}
{$IFDEF UDB_DBX} TSimpleDataSet {$ENDIF}
{$IFDEF UDB_IB} TIBQuery {$ENDIF}
{$IFDEF UDB_FIB} TpFIBDataSet {$ENDIF}
{$IFDEF UDB_FD} TFDQuery {$ENDIF})
  private
    FTableName: string;
    FIdFieldName: string;
    FFieldImgIdx: TStrings;
    FFieldPropsAutoLoad: boolean;
    FFieldDefFormatAutoSet: boolean;
    FFieldTagToDuplic: byte;
    FDuplicBeforeAfterInsert: boolean;
    FFieldTagToReadOnly: byte;
    FConfirmDelete: boolean;
    FCheckFieldSizeBeforePost: boolean;
    FLastSort: string;
    FDataSetEvent: TDataSetEvent;
    FEventNotifyArr: array of TDataSetEvent;
    FControlsToUpdate: TStrings;
    FControlsReadOnlyColor: TColor;
    FControlsUpdating: boolean;
    FLastEditAction: TDatasetLastEditAction;
{$IFDEF UDB_AddQueryReadOnly}
    FReadOnly: boolean;
{$ENDIF}
{$IFDEF UDB_AddQuerySQL}
    FSQL: TStrings;
{$ENDIF}
{$IFDEF UDB_DBX}
    FAutoApplyUpdates: boolean;
    procedure IntApplyUpdates;
    procedure IntReconcileError(DataSet: TCustomClientDataSet; E: EReconcileError; UpdateKind: TUpdateKind; var Action: TReconcileAction);
{$ENDIF}
    procedure SetFieldImgIdx(Value: TStrings);
    procedure SetControlsToUpdate(Value: TStrings);
{$IFDEF UDB_AddQueryReadOnly}
    function GetReadOnly: boolean;
{$ENDIF}
{$IFDEF UDB_AddQuerySQL}
    function GetSQL: TStrings;
    procedure SetSQL(Value: TStrings);
    procedure SQLChanged(Sender: TObject);
{$ENDIF}
{$IFDEF UDB_AddQueryConnection}
    function GetConnection: {$IFDEF UDB_IB} TIBDataBase {$ENDIF} {$IFDEF UDB_FIB} TFIBDatabase {$ENDIF};
    procedure SetConnection(Value: {$IFDEF UDB_IB} TIBDataBase {$ENDIF} {$IFDEF UDB_FIB} TFIBDatabase {$ENDIF});
{$ENDIF}
    procedure AddIntLog(const act: string; AddRecCount: boolean);
    procedure ControlsToUpdateChange(Sender: TObject);
  protected
    procedure Loaded; override;
{$IFDEF UDB_AddQueryReadOnly}
    function GetCanModify: boolean; override;
{$ENDIF}
    procedure DoBeforeOpen; override;
    procedure DoAfterOpen; override;
    procedure DoBeforeClose; override;
    procedure DoAfterClose; override;
    procedure DoBeforePost; override;
    procedure DoAfterPost; override;
    procedure DoBeforeScroll; override;
    procedure DoAfterScroll; override;
    procedure DoBeforeCancel; override;
    procedure DoAfterCancel; override;
    procedure DoBeforeInsert; override;
    procedure DoAfterInsert; override;
    procedure DoBeforeEdit; override;
    procedure DoAfterEdit; override;
    procedure DoBeforeDelete; override;
    procedure DoAfterDelete; override;
{$IFDEF UDB_DBX}
    procedure DoAfterApplyUpdates(var OwnerData: OleVariant); override;
    procedure DoBeforeApplyUpdates(var OwnerData: OleVariant); override;
{$ENDIF}
    procedure CallDatasetEvent(Event: TDatasetEventType);
    procedure ProcControlsToUpdate(Event: TDatasetEventType);
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Post; override; {< @exclude }
    procedure SetReadOnly(Value: boolean);
    procedure SetSortEx(const ASort: string);
    procedure RequeryEx(ChangeCursor: boolean = true);
    procedure SetSQLWhere(const Condition: string); overload;
    procedure SetSQLWhere(const FormatCondition: string; Value: variant); overload;
    procedure SetSQLWhere(Field: TField); overload;
    procedure SetSQLWhere(Field: TField; Value: variant); overload;
    procedure AddSQLWhereOR(const Condition: string);
    procedure AddSQLWhereAND(const Condition: string);
    function SetExecSQL(const SQLCmd: string): integer;
    procedure SetFieldDefaultFormat;
    procedure LoadFieldsProps;
{$IFDEF UDB_Zeos}
    property ResultSet;
    function FindMetadataColumnIndex(const FieldName: string): integer;
{$ENDIF}
    function GetTableName: string;
    procedure UpdateDatasetState;
    function IsEditing: boolean;
    function IsNewRec: boolean;
    function HasFieldToDuplic: boolean;
    function SetFieldsReadOnly: integer;
    function DuplicateRecord: boolean;
    procedure AddEventNotify(CallBackProc: TDataSetEvent);
    procedure RemoveEventNotify(CallBackProc: TDataSetEvent);
    property LastEditAction: TDatasetLastEditAction read FLastEditAction;
  published
    property DataSetField;
    property TableName: string read FTableName write FTableName;
    property IdFieldName: string read FIdFieldName write FIdFieldName;
    property FieldImgIdx: TStrings read FFieldImgIdx write SetFieldImgIdx;
    property FieldPropsAutoLoad: boolean read FFieldPropsAutoLoad write FFieldPropsAutoLoad default true;
    property FieldDefFormatAutoSet: boolean read FFieldDefFormatAutoSet write FFieldDefFormatAutoSet default true;
    property ControlsToUpdate: TStrings read FControlsToUpdate write SetControlsToUpdate;
    property ControlsReadOnlyColor: TColor read FControlsReadOnlyColor write FControlsReadOnlyColor default clBtnFace;
    property ConfirmDelete: boolean read FConfirmDelete write FConfirmDelete default false;
{$IFDEF UDB_AddQueryReadOnly}
    property ReadOnly: boolean read GetReadOnly write SetReadOnly default false;
{$ENDIF}
    property FieldTagToDuplic: byte read FFieldTagToDuplic write FFieldTagToDuplic default 1;
    property DuplicBeforeAfterInsert: boolean read FDuplicBeforeAfterInsert write FDuplicBeforeAfterInsert default false;
    property FieldTagToReadOnly: byte read FFieldTagToReadOnly write FFieldTagToReadOnly default 2;
    property CheckFieldSizeBeforePost: boolean read FCheckFieldSizeBeforePost write FCheckFieldSizeBeforePost default false;
{$IFDEF UDB_AddQuerySQL}
    property SQL: TStrings read GetSQL write SetSQL;
{$ENDIF}
{$IFDEF UDB_DBX}
    property AutoApplyUpdates: boolean read FAutoApplyUpdates write FAutoApplyUpdates default true;
{$ENDIF}
{$IFDEF UDB_AddQueryConnection}
    property Connection: {$IFDEF UDB_IB} TIBDataBase {$ENDIF} {$IFDEF UDB_FIB} TFIBDatabase {$ENDIF} read GetConnection write SetConnection;
{$ENDIF}
    property OnDataSetEvent: TDataSetEvent read FDataSetEvent write FDataSetEvent;
  end;

  { TCustomQueryUDB descendand for internal purpose only }
  { @exclude }
  TQueryIntList = class(TCustomQueryUDB)
  published
    constructor Create(AOwner: TComponent); override; {< @exclude }
  end;

  { TSpeedButton descendand for internal purpose only as bounded button }
  { @exclude }
  TBoundComboButton = class(TSpeedButton)
  private
    ParentW: TControl;
    FIdx: integer;
    GlyphResourceName: string;
    procedure SetIdx(Value: integer);
  protected
    FScaleRatio: double;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF}); override;
    procedure Loaded; override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
  public
    constructor CreateBtn(AOwner: TComponent; AName, ACaption, AGlyphResourceName: string; Idx: integer); virtual;
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    procedure RefreshPosition;
    procedure Click; override;
    property Idx: integer read FIdx write SetIdx;
  published
    property Left stored false;
    property Top stored false;
    property Width stored false;
    property Height stored false;
  end;

  { TrDBCheckBox descendand with boolean value 1 and 0 }
  TCustomrDBCheckBoxUDB = class(TrDBCheckBox)
  public
    constructor Create(AOwner: TComponent); override;
  end;

  { TEdit descendant loads and display text from any SQL command }
  TCustomEditUDB = class(TEdit)
  private
    FListSQL: TStrings;
    FListConnection: TCustomConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListDataSetEvent: TDataSetEvent;
    procedure SetListSQL(Value: TStrings);
    procedure SetListConnection(Value: TCustomConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure CMHintShow(var Message: TMessage); message CM_HINTSHOW;
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh text value }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
  published
    { define SQL command, value of first field is used as text value }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define UDB connection }
    property ListConnection: TCustomConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
  end;

  { TComboBox descendant load list of items from any SQL command }
  TCustomComboBoxUDB = class(TrComboBoxEx)
  private
    FListSQL: TStrings;
    FListConnection: TCustomConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FDelKeyForNullValue: boolean;
    FListDataSetEvent: TDataSetEvent;
    procedure SetListSQL(Value: TStrings);
    procedure SetListConnection(Value: TCustomConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure CMHintShow(var Message: TMessage); message CM_HINTSHOW;
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DoEnter; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected item (record) }
    function GetListField(FieldIndex: integer): TField; overload;
    { return ID value from selected item }
    function GetItemObjectAsInt: integer;
    { set ItemIndex according ID value }
    function SetActiveItem(ObjectAsInt: integer): integer;
  published
    { define SQL command, value of first field is used as list text value,
      value of 2nd field is used as Object value (must be integer) }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define UDB connection }
    property ListConnection: TCustomConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;
    { define whether DEL key set null value }
    property DelKeyForNullValue: boolean read FDelKeyForNullValue write FDelKeyForNullValue default false;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
  end;

  { TrDBEdit descendant show bound button }
  TCustomrDBEditUDB = class(TrDBEdit)
  private
    FEditButton: TBoundComboButton;
    FEditButtonVisible: boolean;
    FEditButtonAllwaysEnabled: boolean;
    procedure SetEditButtonVisible(Value: boolean);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure UpdateState(Sender: TObject); override; {< @exclude }
  published
    { define aux button beside control }
    property EditButton: TBoundComboButton read FEditButton;
    { define visibility of aux button beside control }
    property EditButtonVisible: boolean read FEditButtonVisible write SetEditButtonVisible default true;
    { define enability of aux button beside control }
    property EditButtonAllwaysEnabled: boolean read FEditButtonAllwaysEnabled write FEditButtonAllwaysEnabled default true;
  end;

  { TrDBComboBox descendant load list of items from any SQL command }
  TCustomrDBComboBoxUDB = class(TrDBComboBox)
  private
    FListSQL: TStrings;
    FListConnection: TCustomConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListButton: TBoundComboButton;
    FListButtonVisible: boolean;
    FListButtonAllwaysEnabled: boolean;
    FListButtonTableName: string;
    FClickComboButton: TClickComboButton;
    FListDataSetEvent: TDataSetEvent;
    FListAutoReloadEvents: TDatasetEventTypes;
    FLinkedDSEvent: TList;
    procedure SetListSQL(Value: TStrings);
    function GetListConnection: TCustomConnectionUDB;
    procedure SetListConnection(Value: TCustomConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListButtonVisible(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure ComboButtonClick(Sender: TObject);
    procedure DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure DoEnter; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure UpdateState(Sender: TObject); override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
  published
    { define SQL command, value of first field is used as list text value,
      value of 2nd field is used as Object value (must be integer) }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define list UDB connection, if is not used, connection from DataField is used }
    property ListConnection: TCustomConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;
    { define aux button beside control }
    property ListButton: TBoundComboButton read FListButton;
    { define visibility of aux button beside control }
    property ListButtonVisible: boolean read FListButtonVisible write SetListButtonVisible default true;
    { define enability of aux button beside control }
    property ListButtonAllwaysEnabled: boolean read FListButtonAllwaysEnabled write FListButtonAllwaysEnabled default true;
    { define TableName which should be edited after ListButtonClick (not used by component) }
    property ListButtonTableName: string read FListButtonTableName write FListButtonTableName;
    { define event after ListButtonClick }
    property OnClickComboButton: TClickComboButton read FClickComboButton write FClickComboButton;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
    { define Datasource events when ReloadItems is called automatically }
    property ListAutoReloadEvents: TDatasetEventTypes read FListAutoReloadEvents write FListAutoReloadEvents;
  end;

  { TrDBLookupComboBox descendant load list of items from any SQL command }
  TCustomrDBLookupComboBoxUDB = class(TrDBLookupComboBox)
  private
    FListSQL: TStrings;
    FListConnection: TCustomConnectionUDB;
    FListDataSet: TQueryIntList;
    FListDataSource: TDataSource;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListButton: TBoundComboButton;
    FListButtonVisible: boolean;
    FListButtonAllwaysEnabled: boolean;
    FListButtonTableName: string;
    FClickComboButton: TClickComboButton;
    FListDetailButton: TBoundComboButton;
    FListDetailButtonVisible: boolean;
    FClickComboDetailButton: TClickComboButton;
    FListDataSetEvent: TDataSetEvent;
    FListAutoReloadEvents: TDatasetEventTypes;
    FLinkedDSEvent: TList;
    procedure SetListSQL(Value: TStrings);
    function GetListConnection: TCustomConnectionUDB;
    procedure SetListConnection(Value: TCustomConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListButtonVisible(Value: boolean);
    procedure SetListDetailButtonVisible(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure ComboButtonClick(Sender: TObject);
    function GetListValue: string;
    function GetListDetailButtonEnability: boolean;
    procedure DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure DoEnter; override;
    procedure KeyValueChanged; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure UpdateState(Sender: TObject); override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
    { indicate whether list dataset as active }
    function ListActive: boolean;
  published
    { define SQL command, first field is used as keyvalue field,
      2nd field as list value field }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define list UDB connection, if is not used, connection from DataField is used }
    property ListConnection: TCustomConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;

    { define aux button beside control }
    property ListButton: TBoundComboButton read FListButton;
    { define visibility of aux button beside control }
    property ListButtonVisible: boolean read FListButtonVisible write SetListButtonVisible default true;
    { define enability of aux button beside control }
    property ListButtonAllwaysEnabled: boolean read FListButtonAllwaysEnabled write FListButtonAllwaysEnabled default true;
    { define TableName which should be edited after ListButtonClick (not used by component) }
    property ListButtonTableName: string read FListButtonTableName write FListButtonTableName;
    { define event after ListButtonClick }
    property OnClickComboButton: TClickComboButton read FClickComboButton write FClickComboButton;

    { define 2nd aux button beside control }
    property ListDetailButton: TBoundComboButton read FListDetailButton;
    { define visibility of 2nd aux button beside control }
    property ListDetailButtonVisible: boolean read FListDetailButtonVisible write SetListDetailButtonVisible default true;
    { define event after ListDetailButtonClick }
    property OnClickComboDetailButton: TClickComboButton read FClickComboDetailButton write FClickComboDetailButton;
    { return list value for selected item }
    property ListValue: string read GetListValue;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
    { define Datasource events when ReloadItems is called automatically }
    property ListAutoReloadEvents: TDatasetEventTypes read FListAutoReloadEvents write FListAutoReloadEvents;
  end;

  { TListBox descendant load list of items from any SQL command }
  TCustomListBoxUDB = class(TrListBox)
  private
    FListSQL: TStrings;
    FListConnection: TCustomConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListDataSetEvent: TDataSetEvent;
    procedure SetListSQL(Value: TStrings);
    procedure SetListConnection(Value: TCustomConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DoEnter; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
    { return ID value from selected item }
    function GetItemObjectAsInt: integer;
    { set ItemIndex according ID value }
    function SetActiveItem(ObjectAsInt: integer): integer;
  published
    { define SQL command, value of first field is used as list text value,
      value of 2nd field is used as Object value (must be integer) }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define UDB connection }
    property ListConnection: TCustomConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
  end;

  { TrDBListBox descendant load list of items from any SQL command }
  TCustomrDBListBoxUDB = class(TrDBListBox)
  private
    FListSQL: TStrings;
    FListConnection: TCustomConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListButton: TBoundComboButton;
    FListButtonVisible: boolean;
    FListButtonAllwaysEnabled: boolean;
    FListButtonTableName: string;
    FClickComboButton: TClickComboButton;
    FListDataSetEvent: TDataSetEvent;
    FListAutoReloadEvents: TDatasetEventTypes;
    FLinkedDSEvent: TList;
    procedure SetListSQL(Value: TStrings);
    function GetListConnection: TCustomConnectionUDB;
    procedure SetListConnection(Value: TCustomConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListButtonVisible(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure ComboButtonClick(Sender: TObject);
    procedure DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure DoEnter; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure UpdateState(Sender: TObject); override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
  published
    { define SQL command, value of first field is used as list text value,
      value of 2nd field is used as Object value (must be integer) }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define list UDB connection, if is not used, connection from DataField is used }
    property ListConnection: TCustomConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;

    { define aux button beside control }
    property ListButton: TBoundComboButton read FListButton;
    { define visibility of aux button beside control }
    property ListButtonVisible: boolean read FListButtonVisible write SetListButtonVisible default true;
    { define enability of aux button beside control }
    property ListButtonAllwaysEnabled: boolean read FListButtonAllwaysEnabled write FListButtonAllwaysEnabled default true;
    { define TableName which should be edited after ListButtonClick (not used by component) }
    property ListButtonTableName: string read FListButtonTableName write FListButtonTableName;
    { define event after ListButtonClick }
    property OnClickComboButton: TClickComboButton read FClickComboButton write FClickComboButton;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
    { define Datasource events when ReloadItems is called automatically }
    property ListAutoReloadEvents: TDatasetEventTypes read FListAutoReloadEvents write FListAutoReloadEvents;
  end;

  { TrDBLookupListBox descendant load list of items from any SQL command }
  TCustomrDBLookupListBoxUDB = class(TrDBLookupListBox)
  private
    FListSQL: TStrings;
    FListConnection: TCustomConnectionUDB;
    FListDataSet: TQueryIntList;
    FListDataSource: TDataSource;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    FListButton: TBoundComboButton;
    FListButtonVisible: boolean;
    FListButtonAllwaysEnabled: boolean;
    FListButtonTableName: string;
    FClickComboButton: TClickComboButton;
    FListDataSetEvent: TDataSetEvent;
    FListAutoReloadEvents: TDatasetEventTypes;
    FLinkedDSEvent: TList;
    procedure SetListSQL(Value: TStrings);
    function GetListConnection: TCustomConnectionUDB;
    procedure SetListConnection(Value: TCustomConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListButtonVisible(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
    procedure ComboButtonClick(Sender: TObject);
    function GetListValue: string;
    procedure DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage);
    message CM_ENABLEDCHANGED;
    procedure DoEnter; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure UpdateState(Sender: TObject); override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
  published
    { define SQL command, first field is used as keyvalue field,
      2nd field as list value field }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define list UDB connection, if is not used, connection from DataField is used }
    property ListConnection: TCustomConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;

    { define aux button beside control }
    property ListButton: TBoundComboButton read FListButton;
    { define visibility of aux button beside control }
    property ListButtonVisible: boolean read FListButtonVisible write SetListButtonVisible default true;
    { define enability of aux button beside control }
    property ListButtonAllwaysEnabled: boolean read FListButtonAllwaysEnabled write FListButtonAllwaysEnabled default true;
    { define TableName which should be edited after ListButtonClick (not used by component) }
    property ListButtonTableName: string read FListButtonTableName write FListButtonTableName;
    { define event after ListButtonClick }
    property OnClickComboButton: TClickComboButton read FClickComboButton write FClickComboButton;
    { return list value for selected item }
    property ListValue: string read GetListValue;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
    { define Datasource events when ReloadItems is called automatically }
    property ListAutoReloadEvents: TDatasetEventTypes read FListAutoReloadEvents write FListAutoReloadEvents;
  end;

  { TCheckListBox descendant with possibility to load list of items from any SQL command }
  TCustomCheckListBoxUDB = class(TrCheckListBox)
  private
    FListSQL: TStrings;
    FListConnection: TCustomConnectionUDB;
    FListDataSet: TQueryIntList;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FGetListSQL: TGetSQLEvent;
    procedure SetListSQL(Value: TStrings);
    procedure SetListConnection(Value: TCustomConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DoEnter; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItems;
    { return internal Dataset to allow reading of other fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
    { return ID value from selected item }
    function GetItemObjectAsInt: integer;
    { set ItemIndex according ID value }
    function SetActiveItem(ObjectAsInt: integer): integer;
  published
    { define SQL command, value of first field is used as text value }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define UDB connection }
    property ListConnection: TCustomConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadItems is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FGetListSQL write FGetListSQL;
  end;

  { TrDBCheckListBox descendant with possibility to load list of items and values
    from any SQL command, provide all functions for N:M relationship }
  TCustomrDBCheckListBoxUDB = class(TrDBCheckListBox)
  private
    FListConnection: TCustomConnectionUDB;
    FItemsListDataSet: TQueryIntList;
    FItemsListSQL: TStrings;
    FItemsListAutoLoad: boolean;
    FItemsListSQLTryNow: boolean;
    FGetItemsListSQL: TGetSQLEvent;
    FItemsValueDataSet: TQueryIntList;
    FItemsValueSQL: TStrings;
    FItemsValueSQLTryNow: boolean;
    FGetItemsValueSQL: TGetSQLEvent;
    FItemsValueDataSetEvent: TDataSetEvent;
    FLinkedDSEvent: TList;
    FOnBeforeSaveItemValues: TLoadSaveItemList;
    function GetListConnection: TCustomConnectionUDB;
    procedure SetListConnection(Value: TCustomConnectionUDB);
    procedure SetItemsListSQL(Value: TStrings);
    procedure SetItemsListSQLTryNow(Value: boolean);
    procedure SetItemsValueSQL(Value: TStrings);
    procedure SetItemsValueSQLTryNow(Value: boolean);
    procedure SetItemsValueDataSetEvent(Value: TDataSetEvent);
    procedure DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure DoLoadItemValues; override; {< @exclude }
    procedure DoSaveItemValues; override; {< @exclude }
    { open SQL command and refresh list items }
    procedure ReloadItemsList(ReloadValues: boolean);
    { open SQL command and refresh list values = checkboxes }
    procedure ReloadItemsValue;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(const FieldName: string): TField; overload;
    { return Field from internal Dataset to allow reading of field value from selected record }
    function GetListField(FieldIndex: integer): TField; overload;
    { return ID value from selected item }
    function GetItemObjectAsInt: integer;
    { set ItemIndex according ID value }
    function SetActiveItem(ObjectAsInt: integer): integer;
  published
    { define UDB connection }
    property ListConnection: TCustomConnectionUDB read FListConnection write SetListConnection;
    { define SQL command for loading of list items,
      value of first field is used as N-table ID value, value of 2nd field as N-table listed text }
    property ItemsListSQL: TStrings read FItemsListSQL write SetItemsListSQL;
    { define whether ReloadItemsList is called automatically after loading }
    property ItemsListAutoLoad: boolean read FItemsListAutoLoad write FItemsListAutoLoad default true;
    { try to open ItemsListSQL command in design time }
    property ItemsListSQLTryNow: boolean read FItemsListSQLTryNow write SetItemsListSQLTryNow default false;
    { call before opening of ItemsListSQL to allow update command or set WHERE condition }
    property OnGetItemsListSQL: TGetSQLEvent read FGetItemsListSQL write FGetItemsListSQL;
    { define SQL command for loading of item values,
      value of first field is used as M-table ID value, value of 2nd field as N-table ID value }
    property ItemsValueSQL: TStrings read FItemsValueSQL write SetItemsValueSQL;
    { try to open ItemsValueSQL command in design time }
    property ItemsValueSQLTryNow: boolean read FItemsValueSQLTryNow write SetItemsValueSQLTryNow default false;
    { call before opening of ItemsValueSQL to allow update command or set WHERE condition }
    property OnGetItemsValueSQL: TGetSQLEvent read FGetItemsValueSQL write FGetItemsValueSQL;
    { return internal List Dataset to allow reading of other fields }
    property ItemsListDataSet: TQueryIntList read FItemsListDataSet;
    { return internal Value Dataset to allow reading of other fields }
    property ItemsValueDataSet: TQueryIntList read FItemsValueDataSet;
    { define event called before saving of items }
    property OnBeforeSaveItemValues: TLoadSaveItemList read FOnBeforeSaveItemValues write FOnBeforeSaveItemValues;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnItemsValueDataSetEvent: TDataSetEvent read FItemsValueDataSetEvent write SetItemsValueDataSetEvent;
  end;

  { TrDBRecView descendant with possibility to load/show/edit record from any SQL command }
  TCustomrDBRecViewUDB = class(TrDBRecView)
  private
    FRecSQL: TStrings;
    FRecConnection: TCustomConnectionUDB;
    FRecDataSet: TQueryIntList;
    FRecDataSource: TDataSource;
    FTemplateDataSource: TDataSource;
    FRecDataSetReadOnly: boolean;
    FRecAutoLoad: boolean;
    FRecSQLTryNow: boolean;
    FOnGetRecSQL: TGetSQLEvent;
    FRecDataSetEvent: TDataSetEvent;
    procedure SetRecSQL(Value: TStrings);
    procedure SetRecConnection(Value: TCustomConnectionUDB);
    procedure SetRecSQLTryNow(Value: boolean);
    procedure SetRecDataSetEvent(Value: TDataSetEvent);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh recview content }
    procedure ReloadData;
    { return internal Dataset to allow reading of other fields }
    property RecDataSet: TQueryIntList read FRecDataSet;
    { return internal Datasource to allow using for other DB components }
    property RecDataSource: TDataSource read FRecDataSource;
  published
    { define SQL command }
    property RecSQL: TStrings read FRecSQL write SetRecSQL;
    { define UDB connection }
    property RecConnection: TCustomConnectionUDB read FRecConnection write SetRecConnection;
    { define whether ReloadData is called automatically after loading }
    property RecAutoLoad: boolean read FRecAutoLoad write FRecAutoLoad default true;
    { try to open SQL command in design time }
    property RecSQLTryNow: boolean read FRecSQLTryNow write SetRecSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetRecSQL: TGetSQLEvent read FOnGetRecSQL write FOnGetRecSQL;
    { define whether internal dataset is read only (editing is not allowed) }
    property RecDataSetReadOnly: boolean read FRecDataSetReadOnly write FRecDataSetReadOnly default false;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnRecDataSetEvent: TDataSetEvent read FRecDataSetEvent write SetRecDataSetEvent;
  end;

  { TrDBGrid_MS descendant with possibility to load/show/edit records from any SQL command }
  TCustomrDBGrid_MSUDB = class(TrDBGrid_MS)
  private
    FListSQL: TStrings;
    FListConnection: TCustomConnectionUDB;
    FListDataSet: TQueryIntList;
    FListDataSource: TDataSource;
    FListDataSetReadOnly: boolean;
    FListAutoLoad: boolean;
    FListSQLTryNow: boolean;
    FOnGetListSQL: TGetSQLEvent;
    FListDataSetEvent: TDataSetEvent;
    procedure SetListSQL(Value: TStrings);
    procedure SetListConnection(Value: TCustomConnectionUDB);
    procedure SetListSQLTryNow(Value: boolean);
    procedure SetListDataSetEvent(Value: TDataSetEvent);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { open SQL command and refresh Grid content }
    procedure ReloadData;
    { return internal Dataset to allow reading of fields }
    property ListDataSet: TQueryIntList read FListDataSet;
    { return internal Datasource to allow using for other DB components }
    property ListDataSource: TDataSource read FListDataSource;
  published
    { define SQL command }
    property ListSQL: TStrings read FListSQL write SetListSQL;
    { define UDB connection }
    property ListConnection: TCustomConnectionUDB read FListConnection write SetListConnection;
    { define whether ReloadData is called automatically after loading }
    property ListAutoLoad: boolean read FListAutoLoad write FListAutoLoad default true;
    { try to open SQL command in design time }
    property ListSQLTryNow: boolean read FListSQLTryNow write SetListSQLTryNow default false;
    { call before opening of SQL to allow update command or set WHERE condition }
    property OnGetListSQL: TGetSQLEvent read FOnGetListSQL write FOnGetListSQL;
    { define whether internal dataset is read only (editing is not allowed) }
    property ListDataSetReadOnly: boolean read FListDataSetReadOnly write FListDataSetReadOnly default false;
    { define event for internal dataset (BeforeOpen, BeforePost... After... ses TDataSetEventTypes }
    property OnListDataSetEvent: TDataSetEvent read FListDataSetEvent write SetListDataSetEvent;
  end;

  { universal sorter for any UDB dataset }
  TCustomrDBGridSorter_UDB = class(TrDBGridCustomSorter)
  protected
    function SetSort_CreateIndex(DataSet: TDataSet; const SortString: string; const SortFields: string; const DescendingFields: string): boolean; override;
    procedure DatasetRequery(DataSet: TDataSet); override;
  end;
{$IFDEF UDB_UDBComp}

  { TCustomConnectionUDB compiled for specified dataset type }
  TConnectionUDB = class(TCustomConnectionUDB)
  end;

  { TCustomQueryUDB compiled for specified dataset type }
  TQueryUDB = class(TCustomQueryUDB)
  end;

  { TCustomEditUDB compiled for specified dataset type }
  TEditUDB = class(TCustomEditUDB)
  end;

  { TCustomrDBEditUDB compiled for specified dataset type }
  TrDBEditUDB = class(TCustomrDBEditUDB)
  end;

  { TCustomrDBCheckBoxUDB compiled for specified dataset type }
  TrDBCheckBoxUDB = class(TCustomrDBCheckBoxUDB)
  end;

  { TCustomComboBoxUDB compiled for specified dataset type }
  TComboBoxUDB = class(TCustomComboBoxUDB)
  end;

  { TCustomrDBComboBoxUDB compiled for specified dataset type }
  TrDBComboBoxUDB = class(TCustomrDBComboBoxUDB)
  end;

  { TCustomrDBLookupComboBoxUDB compiled for specified dataset type }
  TrDBLookupComboBoxUDB = class(TCustomrDBLookupComboBoxUDB)
  end;

  { TCustomListBoxUDB compiled for specified dataset type }
  TListBoxUDB = class(TCustomListBoxUDB)
  end;

  { TCustomrDBListBoxUDB compiled for specified dataset type }
  TrDBListBoxUDB = class(TCustomrDBListBoxUDB)
  end;

  { TCustomrDBLookupListBoxUDB compiled for specified dataset type }
  TrDBLookupListBoxUDB = class(TCustomrDBLookupListBoxUDB)
  end;

  { TCustomCheckListBoxUDB compiled for specified dataset type }
  TCheckListBoxUDB = class(TCustomCheckListBoxUDB)
  end;

  { TCustomrDBCheckListBoxUDB compiled for specified dataset type }
  TrDBCheckListBoxUDB = class(TCustomrDBCheckListBoxUDB)
  end;

  { TCustomrDBRecViewUDB compiled for specified dataset type }
  TrDBRecViewUDB = class(TCustomrDBRecViewUDB)
  end;

  { TCustomrDBGrid_MSUDB compiled for specified dataset type }
  TrDBGrid_MSUDB = class(TCustomrDBGrid_MSUDB)
  end;

  { TCustomrDBGridSorter_UDB compiled for specified dataset type }
  TrDBGridSorter_UDB = class(TCustomrDBGridSorter_UDB)
  end;
{$ENDIF}
{$IFDEF UDB_ADOComp}

  TConnectionUDB_ADO = class(TCustomConnectionUDB)
  end;

  TQueryUDB_ADO = class(TCustomQueryUDB)
  end;

  TEditUDB_ADO = class(TCustomEditUDB)
  end;

  TrDBEditUDB_ADO = class(TCustomrDBEditUDB)
  end;

  TrDBCheckBoxUDB_ADO = class(TCustomrDBCheckBoxUDB)
  end;

  TComboBoxUDB_ADO = class(TCustomComboBoxUDB)
  end;

  TrDBComboBoxUDB_ADO = class(TCustomrDBComboBoxUDB)
  end;

  TrDBLookupComboBoxUDB_ADO = class(TCustomrDBLookupComboBoxUDB)
  end;

  TListBoxUDB_ADO = class(TCustomListBoxUDB)
  end;

  TrDBListBoxUDB_ADO = class(TCustomrDBListBoxUDB)
  end;

  TrDBLookupListBoxUDB_ADO = class(TCustomrDBLookupListBoxUDB)
  end;

  TCheckListBoxUDB_ADO = class(TCustomCheckListBoxUDB)
  end;

  TrDBCheckListBoxUDB_ADO = class(TCustomrDBCheckListBoxUDB)
  end;

  TrDBRecViewUDB_ADO = class(TCustomrDBRecViewUDB)
  end;

  TrDBGrid_MSUDB_ADO = class(TCustomrDBGrid_MSUDB)
  end;

  TrDBGridSorter_UDB_ADO = class(TCustomrDBGridSorter_UDB)
  end;
{$ENDIF}
{$IFDEF UDB_DBXComp}

  // specific components
  TConnectionUDB_DBX = class(TCustomConnectionUDB)
  end;

  TQueryUDB_DBX = class(TCustomQueryUDB)
  end;

  TEditUDB_DBX = class(TCustomEditUDB)
  end;

  TrDBEditUDB_DBX = class(TCustomrDBEditUDB)
  end;

  TrDBCheckBoxUDB_DBX = class(TCustomrDBCheckBoxUDB)
  end;

  TComboBoxUDB_DBX = class(TCustomComboBoxUDB)
  end;

  TrDBComboBoxUDB_DBX = class(TCustomrDBComboBoxUDB)
  end;

  TrDBLookupComboBoxUDB_DBX = class(TCustomrDBLookupComboBoxUDB)
  end;

  TListBoxUDB_DBX = class(TCustomListBoxUDB)
  end;

  TrDBListBoxUDB_DBX = class(TCustomrDBListBoxUDB)
  end;

  TrDBLookupListBoxUDB_DBX = class(TCustomrDBLookupListBoxUDB)
  end;

  TCheckListBoxUDB_DBX = class(TCustomCheckListBoxUDB)
  end;

  TrDBCheckListBoxUDB_DBX = class(TCustomrDBCheckListBoxUDB)
  end;

  TrDBRecViewUDB_DBX = class(TCustomrDBRecViewUDB)
  end;

  TrDBGrid_MSUDB_DBX = class(TCustomrDBGrid_MSUDB)
  end;

  TrDBGridSorter_UDB_DBX = class(TCustomrDBGridSorter_UDB)
  end;
{$ENDIF}
{$IFDEF UDB_IBComp}

  // specific components
  TConnectionUDB_IB = class(TCustomConnectionUDB)
  end;

  TQueryUDB_IB = class(TCustomQueryUDB)
  end;

  TEditUDB_IB = class(TCustomEditUDB)
  end;

  TrDBEditUDB_IB = class(TCustomrDBEditUDB)
  end;

  TrDBCheckBoxUDB_IB = class(TCustomrDBCheckBoxUDB)
  end;

  TComboBoxUDB_IB = class(TCustomComboBoxUDB)
  end;

  TrDBComboBoxUDB_IB = class(TCustomrDBComboBoxUDB)
  end;

  TrDBLookupComboBoxUDB_IB = class(TCustomrDBLookupComboBoxUDB)
  end;

  TListBoxUDB_IB = class(TCustomListBoxUDB)
  end;

  TrDBListBoxUDB_IB = class(TCustomrDBListBoxUDB)
  end;

  TrDBLookupListBoxUDB_IB = class(TCustomrDBLookupListBoxUDB)
  end;

  TCheckListBoxUDB_IB = class(TCustomCheckListBoxUDB)
  end;

  TrDBCheckListBoxUDB_IB = class(TCustomrDBCheckListBoxUDB)
  end;

  TrDBRecViewUDB_IB = class(TCustomrDBRecViewUDB)
  end;

  TrDBGrid_MSUDB_IB = class(TCustomrDBGrid_MSUDB)
  end;

  TrDBGridSorter_UDB_IB = class(TCustomrDBGridSorter_UDB)
  end;
{$ENDIF}
{$IFDEF UDB_FIBComp}

  // specific components
  TConnectionUDB_FIB = class(TCustomConnectionUDB)
  end;

  TQueryUDB_FIB = class(TCustomQueryUDB)
  end;

  TEditUDB_FIB = class(TCustomEditUDB)
  end;

  TrDBEditUDB_FIB = class(TCustomrDBEditUDB)
  end;

  TrDBCheckBoxUDB_FIB = class(TCustomrDBCheckBoxUDB)
  end;

  TComboBoxUDB_FIB = class(TCustomComboBoxUDB)
  end;

  TrDBComboBoxUDB_FIB = class(TCustomrDBComboBoxUDB)
  end;

  TrDBLookupComboBoxUDB_FIB = class(TCustomrDBLookupComboBoxUDB)
  end;

  TListBoxUDB_FIB = class(TCustomListBoxUDB)
  end;

  TrDBListBoxUDB_FIB = class(TCustomrDBListBoxUDB)
  end;

  TrDBLookupListBoxUDB_FIB = class(TCustomrDBLookupListBoxUDB)
  end;

  TCheckListBoxUDB_FIB = class(TCustomCheckListBoxUDB)
  end;

  TrDBCheckListBoxUDB_FIB = class(TCustomrDBCheckListBoxUDB)
  end;

  TrDBRecViewUDB_FIB = class(TCustomrDBRecViewUDB)
  end;

  TrDBGrid_MSUDB_FIB = class(TCustomrDBGrid_MSUDB)
  end;

  TrDBGridSorter_UDB_FIB = class(TCustomrDBGridSorter_UDB)
  end;
{$ENDIF}
{$IFDEF UDB_FDComp}

  // specific components
  TConnectionUDB_FD = class(TCustomConnectionUDB)
  end;

  TQueryUDB_FD = class(TCustomQueryUDB)
  end;

  TEditUDB_FD = class(TCustomEditUDB)
  end;

  TrDBEditUDB_FD = class(TCustomrDBEditUDB)
  end;

  TrDBCheckBoxUDB_FD = class(TCustomrDBCheckBoxUDB)
  end;

  TComboBoxUDB_FD = class(TCustomComboBoxUDB)
  end;

  TrDBComboBoxUDB_FD = class(TCustomrDBComboBoxUDB)
  end;

  TrDBLookupComboBoxUDB_FD = class(TCustomrDBLookupComboBoxUDB)
  end;

  TListBoxUDB_FD = class(TCustomListBoxUDB)
  end;

  TrDBListBoxUDB_FD = class(TCustomrDBListBoxUDB)
  end;

  TrDBLookupListBoxUDB_FD = class(TCustomrDBLookupListBoxUDB)
  end;

  TCheckListBoxUDB_FD = class(TCustomCheckListBoxUDB)
  end;

  TrDBCheckListBoxUDB_FD = class(TCustomrDBCheckListBoxUDB)
  end;

  TrDBRecViewUDB_FD = class(TCustomrDBRecViewUDB)
  end;

  TrDBGrid_MSUDB_FD = class(TCustomrDBGrid_MSUDB)
  end;

  TrDBGridSorter_UDB_FD = class(TCustomrDBGridSorter_UDB)
  end;
{$ENDIF}
{$IFDEF UDB_FDComp}

  // specific components
  TConnectionUDB_Zeos = class(TCustomConnectionUDB)
  end;

  TQueryUDB_Zeos = class(TCustomQueryUDB)
  end;

  TEditUDB_Zeos = class(TCustomEditUDB)
  end;

  TrDBEditUDB_Zeos = class(TCustomrDBEditUDB)
  end;

  TrDBCheckBoxUDB_Zeos = class(TCustomrDBCheckBoxUDB)
  end;

  TComboBoxUDB_Zeos = class(TCustomComboBoxUDB)
  end;

  TrDBComboBoxUDB_Zeos = class(TCustomrDBComboBoxUDB)
  end;

  TrDBLookupComboBoxUDB_Zeos = class(TCustomrDBLookupComboBoxUDB)
  end;

  TListBoxUDB_Zeos = class(TCustomListBoxUDB)
  end;

  TrDBListBoxUDB_Zeos = class(TCustomrDBListBoxUDB)
  end;

  TrDBLookupListBoxUDB_Zeos = class(TCustomrDBLookupListBoxUDB)
  end;

  TCheckListBoxUDB_Zeos = class(TCustomCheckListBoxUDB)
  end;

  TrDBCheckListBoxUDB_Zeos = class(TCustomrDBCheckListBoxUDB)
  end;

  TrDBRecViewUDB_Zeos = class(TCustomrDBRecViewUDB)
  end;

  TrDBGrid_MSUDB_Zeos = class(TCustomrDBGrid_MSUDB)
  end;

  TrDBGridSorter_UDB_Zeos = class(TCustomrDBGridSorter_UDB)
  end;
{$ENDIF}

procedure Register;

var
  lastRequeryUDB: TCustomQueryUDB;

implementation

uses RosiCompGlobalCfg, rstring, rdlg, rtool, rdbtool;

const
  resnameDBLOOKUPDETAIL = 'DBLOOKUPDETAIL';
  resnameDBLOOKUPSELECT = 'DBLOOKUPSELECT';

type
  THackWinControl = class(TWinControl)
  end;

var
  lasttime: TTime;

procedure Register;
begin
{$IFDEF UDB_UDBComp}
  RegisterComponents('Rosi UDB', [TConnectionUDB, TQueryUDB, TEditUDB, TrDBEditUDB, TrDBCheckBoxUDB, TComboBoxUDB, TrDBComboBoxUDB, TrDBLookupComboBoxUDB,
    TListBoxUDB, TrDBListBoxUDB, TrDBLookupListBoxUDB, TCheckListBoxUDB, TrDBCheckListBoxUDB, TrDBRecViewUDB, TrDBGrid_MSUDB, TrDBGridSorter_UDB]);
{$ENDIF}
{$IFDEF UDB_ADOComp}
  RegisterComponents('Rosi UDB ADO', [TConnectionUDB_ADO, TQueryUDB_ADO, TEditUDB_ADO, TrDBEditUDB_ADO, TrDBCheckBoxUDB_ADO, TComboBoxUDB_ADO,
    TrDBComboBoxUDB_ADO, TrDBLookupComboBoxUDB_ADO, TListBoxUDB_ADO, TrDBListBoxUDB_ADO, TrDBLookupListBoxUDB_ADO, TCheckListBoxUDB_ADO,
    TrDBCheckListBoxUDB_ADO, TrDBRecViewUDB_ADO, TrDBGrid_MSUDB_ADO, TrDBGridSorter_UDB_ADO]);
{$ENDIF}
{$IFDEF UDB_DBXComp}
  RegisterComponents('Rosi UDB DBX', [TConnectionUDB_DBX, TQueryUDB_DBX, TEditUDB_DBX, TrDBEditUDB_DBX, TrDBCheckBoxUDB_DBX, TComboBoxUDB_DBX,
    TrDBComboBoxUDB_DBX, TrDBLookupComboBoxUDB_DBX, TListBoxUDB_DBX, TrDBListBoxUDB_DBX, TrDBLookupListBoxUDB_DBX, TCheckListBoxUDB_DBX,
    TrDBCheckListBoxUDB_DBX, TrDBRecViewUDB_DBX, TrDBGrid_MSUDB_DBX, TrDBGridSorter_UDB_DBX]);
{$ENDIF}
{$IFDEF UDB_IBComp}
  RegisterComponents('Rosi UDB IB', [TConnectionUDB_IB, TQueryUDB_IB, TEditUDB_IB, TrDBEditUDB_IB, TrDBCheckBoxUDB_IB, TComboBoxUDB_IB, TrDBComboBoxUDB_IB,
    TrDBLookupComboBoxUDB_IB, TListBoxUDB_IB, TrDBListBoxUDB_IB, TrDBLookupListBoxUDB_IB, TCheckListBoxUDB_IB, TrDBCheckListBoxUDB_IB, TrDBRecViewUDB_IB,
    TrDBGrid_MSUDB_IB, TrDBGridSorter_UDB_IB]);
{$ENDIF}
{$IFDEF UDB_FIBComp}
  RegisterComponents('Rosi UDB FIB', [TConnectionUDB_FIB, TQueryUDB_FIB, TEditUDB_FIB, TrDBEditUDB_FIB, TrDBCheckBoxUDB_FIB, TComboBoxUDB_FIB,
    TrDBComboBoxUDB_FIB, TrDBLookupComboBoxUDB_FIB, TListBoxUDB_FIB, TrDBListBoxUDB_FIB, TrDBLookupListBoxUDB_FIB, TCheckListBoxUDB_FIB,
    TrDBCheckListBoxUDB_FIB, TrDBRecViewUDB_FIB, TrDBGrid_MSUDB_FIB, TrDBGridSorter_UDB_FIB]);
{$ENDIF}
{$IFDEF UDB_FDComp}
  RegisterComponents('Rosi UDB FD', [TConnectionUDB_FD, TQueryUDB_FD, TEditUDB_FD, TrDBEditUDB_FD, TrDBCheckBoxUDB_FD, TComboBoxUDB_FD, TrDBComboBoxUDB_FD,
    TrDBLookupComboBoxUDB_FD, TListBoxUDB_FD, TrDBListBoxUDB_FD, TrDBLookupListBoxUDB_FD, TCheckListBoxUDB_FD, TrDBCheckListBoxUDB_FD, TrDBRecViewUDB_FD,
    TrDBGrid_MSUDB_FD, TrDBGridSorter_UDB_FD]);
{$ENDIF}
{$IFDEF UDB_ZeosComp}
  RegisterComponents('Rosi UDB Zeos', [TConnectionUDB_Zeos, TQueryUDB_Zeos, TEditUDB_Zeos, TrDBEditUDB_Zeos, TrDBCheckBoxUDB_Zeos, TComboBoxUDB_Zeos,
    TrDBComboBoxUDB_Zeos, TrDBLookupComboBoxUDB_Zeos, TListBoxUDB_Zeos, TrDBListBoxUDB_Zeos, TrDBLookupListBoxUDB_Zeos, TCheckListBoxUDB_Zeos,
    TrDBCheckListBoxUDB_Zeos, TrDBRecViewUDB_Zeos, TrDBGrid_MSUDB_Zeos, TrDBGridSorter_UDB_Zeos]);
{$ENDIF}
end;

// ==============================================================================
// TConnectionUDBEx
{$IFDEF UDB_AddConReadOnly}

procedure TCustomConnectionUDB.SetReadOnly(Value: boolean);
begin
{$IFDEF UDB_ADO}
  if Value then
    Mode:=cmRead
  else
    Mode:=cmReadWrite;
  FReadOnly:=Value;
{$ENDIF}
{$IF Defined(UDB_DBX) OR Defined(UDB_IB) OR Defined(UDB_FIB) OR Defined(UDB_FD)}
  FReadOnly:=Value;
{$IFEND}
end;
{$ENDIF}
{$IFDEF UDB_AddConHostName}

function TCustomConnectionUDB.GetHostName: string;
const
  ConStrIds: array [1 .. 2] of string = ('Data Source=', 'SERVER=');
var
  s: string;
{$IFDEF UDB_ADO} i, p: integer; {$ENDIF}
begin
{$IFDEF UDB_ADO}
  for i:= Low(ConStrIds) to High(ConStrIds) do
  begin
    p:=AnsiPosEx(ConStrIds[i], FConnectionStringOrig);
    if p > 0 then
    begin
      s:=copy(FConnectionStringOrig, p, Length(FConnectionStringOrig));
      s:=FindStrPart(s, ';', 1, true);
      s:=copy(s, Length(ConStrIds[i]) + 1, Length(s));
      Break;
    end;
  end;
{$ENDIF}
{$IFDEF UDB_DBX}
  if SameText(DriverName, 'Firebird') or SameText(DriverName, 'InterBase') then
    s:=rtool.FindStrPart(Params.Values['Database'], ':', 1, false)
  else
    s:=Params.Values['HostName'];
{$ENDIF}
{$IFDEF UDB_IB}
  s:=rtool.FindStrPart(DatabaseName, ':', 1, false);
{$ENDIF}
{$IFDEF UDB_FIB}
  s:=rtool.FindStrPart(DBName, ':', 1, false);
{$ENDIF}
{$IFDEF UDB_FD}
  s:=Params.Values['Server'];
{$ENDIF}
  Result:=s;
end;
{$ENDIF}

function TCustomConnectionUDB.GetDatabaseName: string;
const
  ConStrIds: array [1 .. 2] of string = ('Initial Catalog=', 'DATABASE=');
var
  s: string;
{$IFDEF UDB_ADO} i, p: integer; {$ENDIF}
begin
{$IFDEF UDB_ADO}
  for i:= Low(ConStrIds) to High(ConStrIds) do
  begin
    p:=AnsiPosEx(ConStrIds[i], FConnectionStringOrig);
    if p > 0 then
    begin
      s:=copy(FConnectionStringOrig, p, Length(FConnectionStringOrig));
      s:=FindStrPart(s, ';', 1, true);
      s:=copy(s, Length(ConStrIds[i]) + 1, Length(s));
      Break;
    end;
  end;
{$ENDIF}
{$IFDEF UDB_DBX}
  if SameText(DriverName, 'Firebird') or SameText(DriverName, 'InterBase') then
    s:=rtool.FindStrPart(Params.Values['Database'], ':', 2, false)
  else
    s:=Params.Values['Database'];
{$ENDIF}
{$IFDEF UDB_IB}
  s:=rtool.FindStrPart(DatabaseName, ':', 2, false);
{$ENDIF}
{$IFDEF UDB_FIB}
  s:=rtool.FindStrPart(DBName, ':', 2, false);
{$ENDIF}
{$IFDEF UDB_FD}
  s:=Params.Values['Database'];
{$ENDIF}
  Result:=s;
end;

procedure TCustomConnectionUDB.CallProcForRegisteredDataset(CallBackProc: TNotifyEvent);
var
  i: integer;
  Current: TDataSet;
begin
{$IF Defined(UDB_ADO) OR Defined(UDB_DBX) OR Defined(UDB_IB) OR Defined(UDB_FD)}
  for i:=0 to DataSetCount - 1 do
  begin
    Current:=self.DataSets[i];
{$IFDEF UDB_DBX}
    if Current.Owner is TCustomQueryUDB then
      Current:=Current.Owner as TCustomQueryUDB;
{$ENDIF}
    try
      CallBackProc(Current);
    except
      // Ignore.
    end;
  end;
{$IFEND}
{$IFDEF UDB_Zeos}
  for i:=0 to FDataSets.Count - 1 do
  begin
    Current:=TDataSet(FDataSets[i]);
    try
      CallBackProc(Current);
    except
      // Ignore.
    end;
  end;
{$ENDIF}
{$IFDEF UDB_FIB}
  for i:=0 to self.FFIBBases.Count - 1 do
  begin
    if FIBBases[i] <> nil then
      if (FIBBases[i].Owner is TDataSet) then
      begin
        Current:=TDataSet(FIBBases[i].Owner);
        try
          CallBackProc(Current);
        except
          // Ignore.
        end;
      end;
  end;
{$ENDIF}
end;
{$IFDEF UDB_Zeos}

procedure TCustomConnectionUDB.Open;
begin
  Connected:=true;
end;

procedure TCustomConnectionUDB.Close;
begin
  Connected:=false;
end;
{$ENDIF}
{$IFDEF UDB_ADO}

procedure TCustomConnectionUDB.DoConnect;
begin
  if (csDesigning in ComponentState) and (FConnectionStringDesign <> '') then
    ConnectionString:=FConnectionStringDesign;
  FConnectionStringOrig:=ConnectionString;
  inherited;
end;
{$ENDIF}

// ==============================================================================
// TCustomQueryUDB
constructor TCustomQueryUDB.Create(AOwner: TComponent);
begin
  inherited;
  FFieldImgIdx:=TStringList.Create;
  FControlsToUpdate:=TStringList.Create;
  FControlsReadOnlyColor:=clBtnFace;
  FControlsUpdating:=false;
  FFieldPropsAutoLoad:=true;
  FFieldDefFormatAutoSet:=true;
  FConfirmDelete:=false;
  FFieldTagToDuplic:=1;
  FDuplicBeforeAfterInsert:=false;
  FFieldTagToReadOnly:=2;
  FLastSort:='';
{$IFDEF UDB_AddQuerySQL}
  FSQL:=TStringList.Create;
  TStringList(FSQL).OnChange:=SQLChanged;
{$ENDIF}
{$IFDEF UDB_DBX}
  FAutoApplyUpdates:=true;
{$ENDIF}
{$IFDEF UDB_FIB}
  AutoCommit:=true;
  AutoUpdateOptions.AutoReWriteSqls:=true;
  AutoUpdateOptions.CanChangeSQLs:=true;
  AutoUpdateOptions.UpdateOnlyModifiedFields:=true;
{$ENDIF}
end;

destructor TCustomQueryUDB.Destroy;
begin
  FreeAndNil(FFieldImgIdx);
  FreeAndNil(FControlsToUpdate);
{$IFDEF UDB_AddQuerySQL}
  FreeAndNil(FSQL);
{$ENDIF}
  inherited;
end;

procedure TCustomQueryUDB.Loaded;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  lasttime:=Now;
  if FFieldDefFormatAutoSet then
    SetFieldDefaultFormat;
  if FFieldPropsAutoLoad then
    LoadFieldsProps;
  AddIntLog('init (loading field props)', false);
end;

procedure TCustomQueryUDB.Post;
begin
  inherited;
end;

procedure TCustomQueryUDB.AddIntLog(const act: string; AddRecCount: boolean);
var
  s, sep: string;
  C: TComponent;
begin
  if Assigned(RosiCompConfig.rDBFieldDefaultsCfg.InternalSQLLog) then
  begin
    sep:=RosiCompConfig.rDBFieldDefaultsCfg.InternalLogSeparator;
    C:=self;
    s:=C.Name;
    while C.Owner <> nil do
    begin
      C:=C.Owner;
      s:=C.Name + '.' + s;
      if (C is TForm) or (C is TDataModule) then
        Break;
    end;
    s:=TimeToStr(Now) + sep + s + sep + act + sep;
    s:=s + FormatFloat('0.0000', (Now - lasttime) * 24 * 60 * 60) + sep;
    if AddRecCount then
    begin
      s:=s + IntToStr(RecordCount) + sep;
      s:=s + IntToStr(RecordSize) + sep;
    end;
    RosiCompConfig.rDBFieldDefaultsCfg.InternalSQLLog.Add(s);
  end;
end;
{$IFDEF UDB_AddQueryReadOnly}

function TCustomQueryUDB.GetCanModify: boolean;
begin
  Result:=not FReadOnly and inherited GetCanModify;
end;

function TCustomQueryUDB.GetReadOnly: boolean;
begin
  Result:=FReadOnly;
end;
{$ENDIF}

procedure TCustomQueryUDB.SetReadOnly(Value: boolean);
begin
{$IFDEF UDB_ADO} FReadOnly:=Value; {$ENDIF}
{$IFDEF UDB_Zeos} ReadOnly:=Value; {$ENDIF}
{$IFDEF UDB_DBX} ReadOnly:=Value; {$ENDIF}
{$IFDEF UDB_IB} FReadOnly:=Value; {$ENDIF}
{$IFDEF UDB_FIB} FReadOnly:=Value; {$ENDIF}
{$IFDEF UDB_FD} FReadOnly:=Value; {$ENDIF}
  UpdateDatasetState;
end;
{$IFDEF UDB_AddQuerySQL}

function TCustomQueryUDB.GetSQL: TStrings;
begin
{$IFDEF UDB_DBX}
  if Assigned(DataSet) then
    FSQL.Text:=DataSet.CommandText;
{$ENDIF}
{$IFDEF UDB_FIB}
  if Assigned(SQLs.SelectSQL) then
    FSQL.Text:=SQLs.SelectSQL.Text;
{$ENDIF}
  Result:=FSQL;
end;

procedure TCustomQueryUDB.SetSQL(Value: TStrings);
begin
  FSQL.Assign(Value);
{$IFDEF UDB_DBX}
  if Assigned(DataSet) then
    DataSet.CommandText:=FSQL.Text;
{$ENDIF}
{$IFDEF UDB_FIB}
  if Assigned(SQLs.SelectSQL) then
    SQLs.SelectSQL.Text:=FSQL.Text;
{$ENDIF}
end;

procedure TCustomQueryUDB.SQLChanged(Sender: TObject);
begin
{$IFDEF UDB_DBX}
  if Assigned(DataSet) and (DataSet.CommandText <> FSQL.Text) then
    DataSet.CommandText:=FSQL.Text;
{$ENDIF}
{$IFDEF UDB_FIB}
  if Assigned(SQLs.SelectSQL) and (SQLs.SelectSQL.Text <> FSQL.Text) then
    SQLs.SelectSQL.Text:=FSQL.Text;
{$ENDIF}
end;
{$ENDIF}
{$IFDEF UDB_AddQueryConnection}

function TCustomQueryUDB.GetConnection: {$IFDEF UDB_IB} TIBDataBase {$ENDIF} {$IFDEF UDB_FIB} TFIBDatabase {$ENDIF};
begin
  Result:=Database;
end;

procedure TCustomQueryUDB.SetConnection(Value: {$IFDEF UDB_IB} TIBDataBase {$ENDIF} {$IFDEF UDB_FIB} TFIBDatabase {$ENDIF});
begin
  Database:=Value;
end;
{$ENDIF}
{$IFDEF UDB_DBX}

procedure TCustomQueryUDB.IntApplyUpdates;
begin
  if FAutoApplyUpdates then
  begin
    if not Assigned(OnReconcileError) then
      OnReconcileError:=IntReconcileError;
    ApplyUpdates(0);
  end;
end;

procedure TCustomQueryUDB.IntReconcileError(DataSet: TCustomClientDataSet; E: EReconcileError; UpdateKind: TUpdateKind; var Action: TReconcileAction);
begin
  ShowErr(rstring.sSaveDataErr, E);
  Action:=raCancel;
  // RequeryEx;
end;
{$ENDIF}

procedure TCustomQueryUDB.DoBeforeOpen;
var
  ro: boolean;
begin
  if not(csDesigning in ComponentState) then
  begin
    if Connection is TCustomConnectionUDB then
      ro:=TCustomConnectionUDB(Connection).FReadOnly
    else
      ro:=false;
    if ro and not ReadOnly then
      ReadOnly:=true;
  end;
  inherited;
  CallDatasetEvent(deBeforeOpen);
end;

procedure TCustomQueryUDB.DoAfterOpen;
begin
  if FLastSort <> '' then
    try
{$IFDEF UDB_ADO} Sort:=FLastSort; {$ENDIF}
{$IFDEF UDB_Zeos} SortedFields:=FLastSort; {$ENDIF}
{$IFDEF UDB_DBX} IndexName:=FLastSort; {$ENDIF}
    except
    end;
{$IFDEF UDB_FIB}
  try
    if FAutoUpdateOptions.UpdateTableName = '' then
      FAutoUpdateOptions.UpdateTableName:=GetTableName;
    if FAutoUpdateOptions.KeyFields = '' then
      FAutoUpdateOptions.KeyFields:=IdFieldName;
  except
  end;
{$ENDIF}
{$IFDEF UDB_FD} IndexName:=FLastSort; {$ENDIF}
  inherited;
  FLastEditAction:=leNone;
  CallDatasetEvent(deAfterOpen);
end;

procedure TCustomQueryUDB.DoBeforeClose;
begin
  if Active then
  begin
{$IFDEF UDB_ADO} FLastSort:=Sort; {$ENDIF}
{$IFDEF UDB_Zeos} FLastSort:=SortedFields; {$ENDIF}
{$IFDEF UDB_DBX} FLastSort:=IndexName; {$ENDIF}
{$IFDEF UDB_FD} FLastSort:=IndexName; {$ENDIF}
  end;
  inherited;
  CallDatasetEvent(deBeforeClose);
end;

procedure TCustomQueryUDB.DoAfterClose;
begin
  inherited;
  FLastEditAction:=leNone;
  CallDatasetEvent(deAfterClose);
end;

procedure TCustomQueryUDB.DoBeforePost;
var
  i: integer;
  F: TField;
begin
  if FCheckFieldSizeBeforePost then
    for i:=0 to self.FieldCount - 1 do
    begin
      F:=self.Fields[i];
      if (F is TStringField) and (Length(F.AsString) > F.Size) then
        F.AsString:=copy(F.AsString, 1, F.Size);
    end;
  inherited;
  CallDatasetEvent(deBeforePost);
end;

procedure TCustomQueryUDB.DoAfterPost;
begin
  inherited;
  CallDatasetEvent(deAfterPost);
{$IFDEF UDB_DBX}
  IntApplyUpdates;
{$ELSE}
  CallDatasetEvent(deAfterPostOrApplyUpdates);
{$ENDIF}
end;

procedure TCustomQueryUDB.DoBeforeScroll;
begin
  inherited;
  CallDatasetEvent(deBeforeScroll);
end;

procedure TCustomQueryUDB.DoAfterScroll;
begin
  inherited;
//  FLastEditAction:=leNone;
  CallDatasetEvent(deAfterScroll);
end;

procedure TCustomQueryUDB.DoBeforeCancel;
begin
  inherited;
  CallDatasetEvent(deBeforeCancel);
end;

procedure TCustomQueryUDB.DoAfterCancel;
begin
  inherited;
  FLastEditAction:=leCancel;
  CallDatasetEvent(deAfterCancel);
end;

procedure TCustomQueryUDB.DoBeforeInsert;
begin
  inherited;
  CallDatasetEvent(deBeforeInsert);
end;

procedure TCustomQueryUDB.DoAfterInsert;
begin
  inherited;
  if FLastEditAction<>leDuplic then
    FLastEditAction:=leInsert;
  CallDatasetEvent(deAfterInsert);
end;

procedure TCustomQueryUDB.DoBeforeEdit;
begin
  inherited;
  CallDatasetEvent(deBeforeEdit);
end;

procedure TCustomQueryUDB.DoAfterEdit;
begin
  inherited;
  FLastEditAction:=leEdit;
  CallDatasetEvent(deAfterEdit);
end;

procedure TCustomQueryUDB.DoBeforeDelete;
begin
  inherited;
  if FConfirmDelete then
    if DlgQ(rstring.sDelRecordQ) <> mrYes then
      Abort;
  CallDatasetEvent(deBeforeDelete);
end;

procedure TCustomQueryUDB.DoAfterDelete;
begin
  inherited;
  FLastEditAction:=leDelete;
  CallDatasetEvent(deAfterDelete);
{$IFDEF UDB_DBX}
  IntApplyUpdates;
{$ELSE}
  CallDatasetEvent(deAfterPostOrApplyUpdates);
{$ENDIF}
end;
{$IFDEF UDB_DBX}

procedure TCustomQueryUDB.DoBeforeApplyUpdates(var OwnerData: OleVariant);
begin
  inherited;
  CallDatasetEvent(deBeforeApplyUpdates);
end;

procedure TCustomQueryUDB.DoAfterApplyUpdates(var OwnerData: OleVariant);
begin
  inherited;
  CallDatasetEvent(deAfterApplyUpdates);
  CallDatasetEvent(deAfterPostOrApplyUpdates);
end;
{$ENDIF}

procedure TCustomQueryUDB.CallDatasetEvent(Event: TDatasetEventType);
var
  i: integer;
begin
  for i:=0 to High(FEventNotifyArr) do
  begin
    if @FEventNotifyArr[i] <> nil then
      FEventNotifyArr[i](self, Event);
  end;
  if Assigned(FDataSetEvent) then
    FDataSetEvent(self, Event);

  if Event in [deAfterOpen, deAfterPost, deAfterCancel, deAfterInsert, deAfterEdit, deAfterDelete, deAfterPostOrApplyUpdates] then
    ProcControlsToUpdate(Event);
end;

procedure TCustomQueryUDB.ProcControlsToUpdate(Event: TDatasetEventType);
var
  i: integer;
  C: TComponent;
  W: TWinControl;
  v: string;
begin
  if Owner is TComponent then
  begin
    C:=Owner as TComponent;
    for i:=0 to C.ComponentCount - 1 do
      if C.Components[i] is TWinControl then
      begin
        W:=C.Components[i] as TWinControl;
        v:=FControlsToUpdate.Values[W.Name];
        if (v <> '') then
        begin
          if AnsiPosEx('EN', v) > 0 then
            W.Enabled:=IsEditing;
          if AnsiPosEx('RO', v) > 0 then
          begin
            if W is TCustomEdit then
              TCustomEdit(W).ReadOnly:=not IsEditing;
          end;
          if AnsiPosEx('COL', v) > 0 then
          begin
            if IsEditing then
              THackWinControl(W).Color:=clWindow
            else
              THackWinControl(W).Color:=FControlsReadOnlyColor;
          end;
          if AnsiPosEx('EDIT', v) > 0 then
          begin
            if W is TEdit then
              TEdit(W).OnChange:=ControlsToUpdateChange;
            if W is TMemo then
              TMemo(W).OnChange:=ControlsToUpdateChange;
            if W is TComboBox then
              TComboBox(W).OnChange:=ControlsToUpdateChange;
            if W is TListBox then
              TListBox(W).OnClick:=ControlsToUpdateChange;
            if W is TCheckBox then
              TCheckBox(W).OnClick:=ControlsToUpdateChange;
            if W is TRadioGroup then
              TRadioGroup(W).OnClick:=ControlsToUpdateChange;
          end;
          if (AnsiPosEx('RESET', v) > 0) and not IsEditing then
          begin
            FControlsUpdating:=true;
            try
              if W is TCustomEdit then
                TCustomEdit(W).Text:='';
            finally
              FControlsUpdating:=false;
            end;
          end;
        end;
      end;
  end;
end;

procedure TCustomQueryUDB.ControlsToUpdateChange(Sender: TObject);
begin
  if FControlsUpdating or IsEditing then
    Exit;
  if GetCanModify then
    Edit;
end;

procedure TCustomQueryUDB.SetSortEx(const ASort: string);
var
  B: TBookmark;
begin
  FLastSort:=ASort;
  if Active then
  begin
    B:=GetBookmark;
    try
{$IFDEF UDB_ADO} Sort:=ASort; {$ENDIF}
{$IFDEF UDB_Zeos} SortedFields:=ASort; {$ENDIF}
{$IFDEF UDB_DBX} IndexName:=ASort; {$ENDIF}
{$IFDEF UDB_FD} IndexName:=ASort; {$ENDIF}
      GotoBookmark(B);
    finally
      FreeBookmark(B);
    end;
  end;
end;

procedure TCustomQueryUDB.SetFieldImgIdx(Value: TStrings);
begin
  FFieldImgIdx.Assign(Value);
end;

procedure TCustomQueryUDB.SetControlsToUpdate(Value: TStrings);
begin
  FControlsToUpdate.Assign(Value);
end;

function TCustomQueryUDB.HasFieldToDuplic: boolean;
var
  i: integer;
begin
  Result:=false;
  if FFieldTagToDuplic = 0 then
    Exit;
  for i:=0 to FieldCount - 1 do
    if (Fields[i].Tag and FFieldTagToDuplic) > 0 then
    begin
      Result:=true;
      Exit;
    end;
end;

function TCustomQueryUDB.SetFieldsReadOnly: integer;
var
  i: integer;
begin
  Result:=0;
  if FFieldTagToReadOnly = 0 then
    Exit;
  for i:=0 to FieldCount - 1 do
  begin
    Fields[i].ReadOnly:=(Fields[i].Tag and FFieldTagToReadOnly) > 0;
    if Fields[i].ReadOnly then
      Inc(Result);
  end;
end;

function TCustomQueryUDB.DuplicateRecord: boolean;
begin
  Result:=FFieldTagToDuplic > 0;
  if Result then
  begin
    FLastEditAction:=leDuplic;
    rdbtool.DuplicRecord(self, FFieldTagToDuplic, FDuplicBeforeAfterInsert);
  end;
end;

procedure TCustomQueryUDB.AddEventNotify(CallBackProc: TDataSetEvent);
var
  i, max: integer;
begin
  max:=Length(FEventNotifyArr);
  { for I:=0 to max - 1 do
    if @FEventNotifyArr[I] = Addr(CallBackProc)^ then
    Exit;
    }
  for i:=0 to max - 1 do
    if @FEventNotifyArr[i] = nil then
    begin
      FEventNotifyArr[i]:=CallBackProc;
      Exit;
    end;
  SetLength(FEventNotifyArr, max + 1);
  FEventNotifyArr[max]:=CallBackProc;
end;

procedure TCustomQueryUDB.RemoveEventNotify(CallBackProc: TDataSetEvent);
var
  i, max: integer;
begin
  max:=Length(FEventNotifyArr);
  for i:=0 to max - 1 do
    if @FEventNotifyArr[i] = @CallBackProc then
      FEventNotifyArr[i]:=nil;
end;

procedure TCustomQueryUDB.RequeryEx(ChangeCursor: boolean = true);
var
  B: TBookmark;
  oldCursor: TCursor;
  id: variant;
begin
  lastRequeryUDB:=self;
  if ChangeCursor then
    oldCursor:=ChCursor(RosiCompConfig.rDBComponentsCfg.RequeryCursorType)
  else
    oldCursor:=Screen.Cursor;
  try
    try
      lasttime:=Now;
      if Active then
      begin
        id:=Fields[0].AsVariant;
        B:=GetBookmark;
        DisableControls;
        try
{$IFDEF UDB_ADO} Requery; {$ENDIF}
{$IFDEF UDB_Zeos} Close; Open; {$ENDIF}// BUG - RereadRows do not work, must Close and Open
{$IFDEF UDB_DBX} Close; Open; {$ENDIF}
{$IFDEF UDB_IB} Requery; {$ENDIF}
{$IFDEF UDB_FIB} Close; Open; {$ENDIF}
{$IFDEF UDB_FD} Close; Open; {$ENDIF}
          try
            GotoBookmark(B);
          except
            try
              if not VarIsNull(id) then
                Locate(Fields[0].FieldName, id, []);
            except
            end;
          end;
        finally
          EnableControls;
        end;
      end
      else
        Open;

      if ChangeCursor then
        AddIntLog('opening (W)', true)
      else
        AddIntLog('opening', true);
    except
      AddIntLog(' ERROR opening ' + rtool.ReplaceStr(#13#10, ' ', self.SQL.Text), false);
      raise ;
    end;
  finally
    if ChangeCursor then
      ChCursor(oldCursor);
  end;
end;

procedure TCustomQueryUDB.SetSQLWhere(const Condition: string);
begin
  rdbtool.SetSQLWhere(SQL, Condition);
end;

procedure TCustomQueryUDB.SetSQLWhere(const FormatCondition: string; Value: variant);
begin
  rdbtool.SetSQLWhere(SQL, FormatCondition, Value);
end;

procedure TCustomQueryUDB.SetSQLWhere(Field: TField);
begin
  rdbtool.SetSQLWhere(SQL, Field);
end;

procedure TCustomQueryUDB.SetSQLWhere(Field: TField; Value: variant);
begin
  rdbtool.SetSQLWhere(SQL, Field, Value);
end;

procedure TCustomQueryUDB.AddSQLWhereOR(const Condition: string);
begin
  rdbtool.AddSQLWhereOR(SQL, Condition);
end;

procedure TCustomQueryUDB.AddSQLWhereAND(const Condition: string);
begin
  rdbtool.AddSQLWhereAND(SQL, Condition);
end;

function TCustomQueryUDB.SetExecSQL(const SQLCmd: string): integer;
begin
  Close;
  SQL.Text:=SQLCmd;
{$IFDEF UDB_ADO} Result:=ExecSQL; {$ENDIF}
{$IFDEF UDB_Zeos} Result:=0; ExecSQL; {$ENDIF}
{$IFDEF UDB_DBX} Result:=0; Execute; {$ENDIF}
{$IFDEF UDB_IB} Result:=0; ExecSQL; {$ENDIF}
{$IFDEF UDB_FIB} Result:=0; Open; {$ENDIF}
{$IFDEF UDB_FD} Result:=0; ExecSQL; {$ENDIF}
  Close;
end;

procedure TCustomQueryUDB.SetFieldDefaultFormat;
var
  i: integer;
  F: TField;
begin
  for i:=0 to self.FieldCount - 1 do
  begin
    F:=self.Fields[i];
    if F is TFloatField then
    begin
      if TFloatField(F).DisplayFormat = '' then
        TFloatField(F).DisplayFormat:=RosiCompConfig.rDBFieldDefaultsCfg.DefaultFloatDisplayFormat;
      if TFloatField(F).EditFormat = '' then
        TFloatField(F).EditFormat:=RosiCompConfig.rDBFieldDefaultsCfg.DefaultFloatEditFormat;
    end;
    { if F is TBCDField then
      begin
      if TBCDField(F).DisplayFormat='' then
      TBCDField(F).DisplayFormat:=DefaultFloatDisplayFormat;
      if TBCDField(F).EditFormat='' then
      TBCDField(F).EditFormat:=DefaultFloatEditFormat;
      end;
      }
    if F is TBooleanField then
    begin
      if TBooleanField(F).DisplayValues = '' then
        TBooleanField(F).DisplayValues:=RosiCompConfig.rDBFieldDefaultsCfg.DefaultBoolDisplayFormat;
    end;
    if F is TDateTimeField then
    begin
      if TDateTimeField(F).DisplayFormat = '' then
      begin
        TDateTimeField(F).DisplayFormat:=RosiCompConfig.rDBFieldDefaultsCfg.DefaultDateTimeDisplayFormat;
        if SameText(copy(F.FieldName, 1, 4), 'Time') then
          TDateTimeField(F).DisplayFormat:=RosiCompConfig.rDBFieldDefaultsCfg.DefaultTimeDisplayFormat;
      end;
    end;
  end;
end;

procedure TCustomQueryUDB.LoadFieldsProps;
begin
  if Assigned(RosiCompConfig.rDBFieldDefaultsCfg.LoadFieldsPropsProc) then
    RosiCompConfig.rDBFieldDefaultsCfg.LoadFieldsPropsProc(self);
end;
{$IFDEF UDB_Zeos}

function TCustomQueryUDB.FindMetadataColumnIndex(const FieldName: string): integer;
var
  i: integer;
begin
  Result:=-1;
  if Assigned(ResultSet) and Assigned(ResultSet.GetMetadata) then
  begin
    for i:=1 to ResultSet.GetMetadata.GetColumnCount do
      if SameText(ResultSet.GetMetadata.GetColumnName(i), FieldName) then
      begin
        Result:=i; Break;
      end;
  end;
end;
{$ENDIF}

function TCustomQueryUDB.GetTableName: string;
begin
  Result:=self.TableName;
  if Result = '' then
    Result:=rdbtool.GetSQLTableName(self.SQL);
end;

procedure TCustomQueryUDB.UpdateDatasetState;
begin
  DataEvent(deDataSetChange, 0);
  // DataEvent(deFieldChange, Longint(Self));
end;

function TCustomQueryUDB.IsEditing: boolean;
begin
  Result:=(State in [dsEdit, dsInsert]);
end;

function TCustomQueryUDB.IsNewRec: boolean;
begin
  Result:=(State in [dsInsert]);
end;

constructor TQueryIntList.Create(AOwner: TComponent);
begin
  inherited;
  FFieldPropsAutoLoad:=false;
  FFieldDefFormatAutoSet:=false;
end;

// ==============================================================================
// TBoundComboButton

constructor TBoundComboButton.CreateBtn(AOwner: TComponent; AName, ACaption, AGlyphResourceName: string; Idx: integer);
begin
  inherited Create(AOwner);
  FScaleRatio:=1;
  SetSubComponent(true);
  if AOwner is TControl then
    ParentW:=AOwner as TControl;
  if AName = '' then
    AName:='ComboButton';
  Name:=AName;
  Caption:='';
  FIdx:=Idx;
  GlyphResourceName:=AGlyphResourceName;
  if GlyphResourceName <> '' then
  begin
{$IFDEF DoNotUseGlyphInDesign}
    if (csDesigning in ComponentState) then
      Caption:=IntToStr(FIdx)
    else
{$ENDIF}
    begin
      Glyph.Assign(rGlyphs.BmpFromIndex(rGlyphs.ButtonIndex(GlyphResourceName), FScaleRatio));
    end;
  end
  else
    Caption:=ACaption;
end;

procedure TBoundComboButton.Loaded;
begin
  inherited;
{$IFDEF DoNotUseGlyphInDesign}
  if (csDesigning in ComponentState) then
    Glyph.Width:=0;
{$ENDIF}
  if Glyph.Width > 0 then
    Caption:='';
end;

procedure TBoundComboButton.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  if not(csDesigning in ComponentState) then
    Glyph.Assign(rGlyphs.BmpFromIndex(rGlyphs.ButtonIndex(GlyphResourceName), FScaleRatio));
end;

procedure TBoundComboButton.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  RefreshPosition;
end;

procedure TBoundComboButton.SetIdx(Value: integer);
begin
  FIdx:=Value;
  RefreshPosition;
end;

procedure TBoundComboButton.RefreshPosition;
var
  Size: integer;
begin
  if Assigned(ParentW) then
  begin
    Size:=ParentW.Height - 2;
    if Size > 24 then
      Size:=24;
    if Visible then
      SetBounds(ParentW.Left + ParentW.Width + 2 + (FIdx - 1) * Size, ParentW.Top + 1, Size, Size)
    else
      SetBounds(ParentW.Left + ParentW.Width - Size, ParentW.Top + 1, Size, Size);
  end;
end;

procedure TBoundComboButton.Click;
begin
  if Assigned(Owner) and (Owner is TWinControl) then
    try
      TWinControl(Owner).SetFocus;
    except
    end;
  inherited;
end;

// ==============================================================================
// TCheckBox

constructor TCustomrDBCheckBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  self.ValueChecked:='1';
  self.ValueUnchecked:='0';
end;

// ==============================================================================
// TEdit

constructor TCustomEditUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListAutoLoad:=true;
  Color:=clBtnFace;
end;

destructor TCustomEditUDB.Destroy;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  inherited;
end;

procedure TCustomEditUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

procedure TCustomEditUDB.SetListConnection(Value: TCustomConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TCustomEditUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TCustomEditUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TCustomEditUDB.CMHintShow(var Message: TMessage);
var
  hnt: string;
  ACanvas: TControlCanvas;
begin
  if Hint = '' then
  begin
    hnt:=Text;
    ACanvas:=TControlCanvas.Create;
    try
      ACanvas.Control:=self;
      ACanvas.Font.Assign(Font);
      if ACanvas.TextWidth(hnt) > self.ClientWidth - 4 then
        TCMHintShow(Message).HintInfo^.HintStr:=hnt;
    finally
      FreeAndNil(ACanvas);
    end;
  end;
  inherited;
end;

procedure TCustomEditUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
  ReadOnly:=true;
end;

procedure TCustomEditUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
end;

procedure TCustomEditUDB.ReloadItems;
var
  F: TField;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=FListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      F:=FListDataSet.Fields[0];
      if IsMemoField(F) then
        self.Text:=F.AsString
      else
        self.Text:=F.DisplayText;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

// ==============================================================================
// TComboBox

constructor TCustomComboBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListAutoLoad:=true;
end;

destructor TCustomComboBoxUDB.Destroy;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  inherited;
end;

procedure TCustomComboBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

procedure TCustomComboBoxUDB.SetListConnection(Value: TCustomConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TCustomComboBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TCustomComboBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TCustomComboBoxUDB.CMHintShow(var Message: TMessage);
var
  hnt: string;
begin
  if Hint = '' then
  begin
    hnt:=Text;
    if Canvas.TextWidth(hnt) > self.ClientWidth - 24 then
      TCMHintShow(Message).HintInfo^.HintStr:=hnt;
  end;
  inherited;
end;

procedure TCustomComboBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TCustomComboBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
end;

procedure TCustomComboBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TCustomComboBoxUDB.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
  if FDelKeyForNullValue and (Style = csDropDownList) and (Key = VK_DELETE) then
  begin
    ItemIndex:=-1;
    Click;
  end;
end;

procedure TCustomComboBoxUDB.ReloadItems;
var
  id: integer;
  s: string;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=FListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      Items.BeginUpdate;
      try
        Items.Clear;
        id:=0;
        while not FListDataSet.Eof do
        begin
          if FListDataSet.FieldCount > 1 then
            id:=FListDataSet.Fields[1].AsInteger;
          s:=FListDataSet.Fields[0].AsString;
          if MaxLength > 0 then
            s:=copy(s, 1, MaxLength);
          Items.AddObject(s, TObject(id));
          FListDataSet.Next;
        end;
      finally
        Items.EndUpdate;
      end;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TCustomComboBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TCustomComboBoxUDB.GetListField(FieldIndex: integer): TField;
var
  i: integer;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  FListDataSet.First;
  for i:=0 to ItemIndex - 1 do
    FListDataSet.Next;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

function TCustomComboBoxUDB.GetItemObjectAsInt: integer;
begin
  if (ItemIndex >= 0) and Assigned(Items.Objects[ItemIndex]) then
    Result:=integer(Items.Objects[ItemIndex])
  else
    Result:=0;
end;

function TCustomComboBoxUDB.SetActiveItem(ObjectAsInt: integer): integer;
begin
  Result:=Items.IndexOfObject(TObject(ObjectAsInt));
  if Result>=0 then
  begin
    ItemIndex:=Result;
    if Assigned(OnClick) then
      OnClick(Self);
  end;
end;

// ==============================================================================
// TCustomrDBEditUDB

constructor TCustomrDBEditUDB.Create(AOwner: TComponent);
begin
  inherited;
  FEditButton:=TBoundComboButton.CreateBtn(self, '', '...', '', 1);
  FEditButton.FreeNotification(self);
  // FEditButton.OnClick:=ComboButtonClick;
  FEditButtonVisible:=true;
  FEditButtonAllwaysEnabled:=true;
end;

destructor TCustomrDBEditUDB.Destroy;
begin
  FreeAndNil(FEditButton);
  inherited;
end;

procedure TCustomrDBEditUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FEditButton) and (Operation = opRemove) then
    FEditButton:=nil;
end;

procedure TCustomrDBEditUDB.SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FEditButton) then
    FEditButton.RefreshPosition;
end;

procedure TCustomrDBEditUDB.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FEditButton) then
    FEditButton.Parent:=AParent;
end;

procedure TCustomrDBEditUDB.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FEditButton) then
    FEditButton.Enabled:=Enabled;
end;

procedure TCustomrDBEditUDB.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FEditButton) then
    FEditButton.Visible:=Visible and FEditButtonVisible;
end;

procedure TCustomrDBEditUDB.SetEditButtonVisible(Value: boolean);
begin
  FEditButtonVisible:=Value;
  if Assigned(FEditButton) then
    FEditButton.Visible:=Visible and FEditButtonVisible;
end;

procedure TCustomrDBEditUDB.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  ro:=ReadOnly or (Assigned(Field) and Field.ReadOnly);
  if Assigned(FEditButton) then
    FEditButton.Enabled:=Enabled and not ro and (FEditButtonAllwaysEnabled or (FDataLink.Active and FDataLink.Editing));
end;

// ==============================================================================
// TrDBComboBox

constructor TCustomrDBComboBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);

  FListButton:=TBoundComboButton.CreateBtn(self, '', '', resnameDBLOOKUPSELECT, 1);
  FListButton.FreeNotification(self);
  FListButton.OnClick:=ComboButtonClick;

  FLinkedDSEvent:=TList.Create;

  FListAutoLoad:=true;
  FListButtonVisible:=true;
  FListButtonAllwaysEnabled:=true;
end;

destructor TCustomrDBComboBoxUDB.Destroy;
var
  i: integer;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  FreeAndNil(FListButton);
  for i:=0 to FLinkedDSEvent.Count - 1 do
    try
      TCustomQueryUDB(FLinkedDSEvent[i]).RemoveEventNotify(DSEventCallBack);
    except
    end;
  FreeAndNil(FLinkedDSEvent);
  inherited;
end;

procedure TCustomrDBComboBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TCustomrDBComboBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
  if (AComponent = FListButton) and (Operation = opRemove) then
    FListButton:=nil;
end;

procedure TCustomrDBComboBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TCustomrDBComboBoxUDB.SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FListButton) then
    FListButton.RefreshPosition;
end;

procedure TCustomrDBComboBoxUDB.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FListButton) then
    FListButton.Parent:=AParent;
end;

procedure TCustomrDBComboBoxUDB.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled;
end;

procedure TCustomrDBComboBoxUDB.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TCustomrDBComboBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

function TCustomrDBComboBoxUDB.GetListConnection: TCustomConnectionUDB;
begin
  if Assigned(FListConnection) then
    Result:=FListConnection
  else
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
      Result:=TCustomConnectionUDB(TCustomQueryUDB(DataSource.DataSet).Connection)
    else
      Result:=nil;
end;

procedure TCustomrDBComboBoxUDB.SetListConnection(Value: TCustomConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TCustomrDBComboBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TCustomrDBComboBoxUDB.SetListButtonVisible(Value: boolean);
begin
  FListButtonVisible:=Value;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TCustomrDBComboBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TCustomrDBComboBoxUDB.ComboButtonClick(Sender: TObject);
var
  tbl: string;
  F: TField;
begin
  tbl:=ListButtonTableName;
  if tbl = '' then
    tbl:=rdbtool.GetSQLTableName(ListSQL);
  F:=nil;
  if FListDataSet.Active then
    F:=FListDataSet.Fields[0];
  if Assigned(FClickComboButton) then
    FClickComboButton(self, GetListConnection, tbl, Field, F);
end;

procedure TCustomrDBComboBoxUDB.DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
begin
  if EventType in FListAutoReloadEvents then
    if Assigned(DataSource) and (DataSet = DataSource.DataSet) then
      ReloadItems;
end;

procedure TCustomrDBComboBoxUDB.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  if (FListAutoReloadEvents <> []) and Assigned(DataSource) and Assigned(DataSource.DataSet) and (DataSource.DataSet is TCustomQueryUDB) then
    if FLinkedDSEvent.IndexOf(DataSource.DataSet) < 0 then
    begin
      FLinkedDSEvent.Add(DataSource.DataSet);
      TCustomQueryUDB(DataSource.DataSet).AddEventNotify(DSEventCallBack);
    end;
  ro:=ReadOnly or (Assigned(Field) and Field.ReadOnly);
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled and not ro and (FListButtonAllwaysEnabled or (FDataLink.Active and FDataLink.Editing));
end;

procedure TCustomrDBComboBoxUDB.ReloadItems;
var
  id: integer;
  s: string;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=GetListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      Items.BeginUpdate;
      try
        Items.Clear;
        id:=0;
        while not FListDataSet.Eof do
        begin
          if (FListDataSet.FieldCount > 1) and (FListDataSet.Fields[1] is TNumericField) then
            id:=FListDataSet.Fields[1].AsInteger;
          s:=FListDataSet.Fields[0].AsString;
          if MaxLength > 0 then
            s:=copy(s, 1, MaxLength);
          Items.AddObject(s, TObject(id));
          FListDataSet.Next;
        end;
      finally
        Items.EndUpdate;
      end;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
    if Assigned(Field) and Field.DataSet.Active then
      try
        ItemIndex:=Items.IndexOf(Field.AsString);
      except
      end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TCustomrDBComboBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TCustomrDBComboBoxUDB.GetListField(FieldIndex: integer): TField;
var
  F: TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  F:=FDataLink.DataSet.FindField(DataField);
  if Assigned(F) then
    rdbtool.LocateRecord(FListDataSet.Fields[0], F.AsString);
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

// ==============================================================================
// TrDBLookupComboBox

constructor TCustomrDBLookupComboBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListDataSource:=TDataSource.Create(self);

  FListButton:=TBoundComboButton.CreateBtn(self, '', '', resnameDBLOOKUPSELECT, 1);
  FListButton.FreeNotification(self);
  FListButton.OnClick:=ComboButtonClick;

  FListDetailButton:=TBoundComboButton.CreateBtn(self, 'ComboDetailButton', '', resnameDBLOOKUPDETAIL, 2);
  FListDetailButton.FreeNotification(self);
  FListDetailButton.OnClick:=ComboButtonClick;

  FLinkedDSEvent:=TList.Create;

  FListAutoLoad:=true;
  FListButtonVisible:=true;
  FListButtonAllwaysEnabled:=true;
  FListDetailButtonVisible:=true;
end;

destructor TCustomrDBLookupComboBoxUDB.Destroy;
var
  i: integer;
begin
  FListDataSet.Close;
  FListDataSource.DataSet:=nil;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  FreeAndNil(FListDataSource);
  FreeAndNil(FListButton);
  FreeAndNil(FListDetailButton);
  for i:=0 to FLinkedDSEvent.Count - 1 do
    try
      TCustomQueryUDB(FLinkedDSEvent[i]).RemoveEventNotify(DSEventCallBack);
    except
    end;
  FreeAndNil(FLinkedDSEvent);
  inherited;
end;

procedure TCustomrDBLookupComboBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TCustomrDBLookupComboBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
  if (AComponent = FListButton) and (Operation = opRemove) then
    FListButton:=nil;
  if (AComponent = FListDetailButton) and (Operation = opRemove) then
    FListDetailButton:=nil;
end;

procedure TCustomrDBLookupComboBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

function TCustomrDBLookupComboBoxUDB.GetListDetailButtonEnability: boolean;
begin
  Result:=Enabled and not VarIsNull(KeyValue) and (not Assigned(Field) or not Field.IsNull);
end;

procedure TCustomrDBLookupComboBoxUDB.KeyValueChanged;
begin
  inherited;
  if Assigned(FListDetailButton) then
    FListDetailButton.Enabled:=GetListDetailButtonEnability;
end;

procedure TCustomrDBLookupComboBoxUDB.SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FListButton) then
    FListButton.RefreshPosition;
  if Assigned(FListDetailButton) then
    FListDetailButton.RefreshPosition;
end;

procedure TCustomrDBLookupComboBoxUDB.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FListButton) then
    FListButton.Parent:=AParent;
  if Assigned(FListDetailButton) then
    FListDetailButton.Parent:=AParent;
end;

procedure TCustomrDBLookupComboBoxUDB.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled;
  if Assigned(FListDetailButton) then
    FListDetailButton.Enabled:=GetListDetailButtonEnability;
end;

procedure TCustomrDBLookupComboBoxUDB.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
  if Assigned(FListDetailButton) then
    FListDetailButton.Visible:=Visible and FListDetailButtonVisible;
end;

procedure TCustomrDBLookupComboBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

function TCustomrDBLookupComboBoxUDB.GetListConnection: TCustomConnectionUDB;
begin
  if Assigned(FListConnection) then
    Result:=FListConnection
  else
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
      Result:=TCustomConnectionUDB(TCustomQueryUDB(DataSource.DataSet).Connection)
    else
      Result:=nil;
end;

procedure TCustomrDBLookupComboBoxUDB.SetListConnection(Value: TCustomConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TCustomrDBLookupComboBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TCustomrDBLookupComboBoxUDB.SetListButtonVisible(Value: boolean);
begin
  FListButtonVisible:=Value;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
  if Assigned(FListDetailButton) then
    if FListButtonVisible then
      FListDetailButton.Idx:=2
    else
      FListDetailButton.Idx:=1;
end;

procedure TCustomrDBLookupComboBoxUDB.SetListDetailButtonVisible(Value: boolean);
begin
  FListDetailButtonVisible:=Value;
  if Assigned(FListDetailButton) then
    FListDetailButton.Visible:=Visible and FListDetailButtonVisible;
end;

procedure TCustomrDBLookupComboBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TCustomrDBLookupComboBoxUDB.ComboButtonClick(Sender: TObject);
var
  tbl: string;
  F: TField;
begin
  tbl:=ListButtonTableName;
  if tbl = '' then
    tbl:=rdbtool.GetSQLTableName(ListSQL);
  F:=nil;
  if Assigned(ListSource) and Assigned(ListSource.DataSet) and ListSource.DataSet.Active then
    F:=ListSource.DataSet.FieldByName(KeyField);
  if (Sender = FListButton) and Assigned(FClickComboButton) then
    FClickComboButton(self, GetListConnection, tbl, Field, F);
  if (Sender = FListDetailButton) and Assigned(FClickComboDetailButton) then
    FClickComboDetailButton(self, GetListConnection, tbl, Field, F);
end;

procedure TCustomrDBLookupComboBoxUDB.DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
begin
  if EventType in FListAutoReloadEvents then
    if Assigned(DataSource) and (DataSet = DataSource.DataSet) then
      ReloadItems;
end;

procedure TCustomrDBLookupComboBoxUDB.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  // reload items after reconnecting
  try
    if FListAutoLoad and (FListDataSource.DataSet = FListDataSet) and not(FListDataSet.Active) and (FListDataSet.Connection.Connected) then
      ReloadItems;
  except
  end;

  if (FListAutoReloadEvents <> []) and Assigned(DataSource) and Assigned(DataSource.DataSet) and (DataSource.DataSet is TCustomQueryUDB) then
    if FLinkedDSEvent.IndexOf(DataSource.DataSet) < 0 then
    begin
      FLinkedDSEvent.Add(DataSource.DataSet);
      TCustomQueryUDB(DataSource.DataSet).AddEventNotify(DSEventCallBack);
    end;
  ro:=ReadOnly or (Assigned(Field) and Field.ReadOnly);
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled and not ro and (FListButtonAllwaysEnabled or (FDataLink.Active and FDataLink.Editing));
  if Assigned(FListDetailButton) then
    FListDetailButton.Enabled:=GetListDetailButtonEnability;
end;

procedure TCustomrDBLookupComboBoxUDB.ReloadItems;
begin
  try
    if (FListSQL.Count = 0) then
    begin
      // standard ListSource component is used
      if Assigned(ListSource) then
      begin
        ListSource.DataSet.Close;
        ListSource.DataSet.Open;
      end;
      Exit;
    end;
    // internal SQL is used
    FListDataSet.Close;
    FListDataSet.Connection:=GetListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      FListDataSet.LoadFieldsProps;
      FListDataSource.DataSet:=FListDataSet;
      ListSource:=FListDataSource;
      if KeyField = '' then
        KeyField:=FListDataSet.Fields[0].FieldName;
      if ListField = '' then
        ListField:=FListDataSet.Fields[1].FieldName;
      DBEditLabel.SetLabelField(ListSource, ListField);
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TCustomrDBLookupComboBoxUDB.GetListValue: string;
var
  F: TField;
begin
  F:=FListDataSet.FindField(ListField);
  if Assigned(F) and not VarIsNull(KeyValue) then
    Result:=F.AsString
  else
    Result:='';
end;

function TCustomrDBLookupComboBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TCustomrDBLookupComboBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

function TCustomrDBLookupComboBoxUDB.ListActive: boolean;
begin
  Result:=Assigned(FListDataSet) and FListDataSet.Active;
end;

// ==============================================================================
// TListBox

constructor TCustomListBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListAutoLoad:=true;
end;

destructor TCustomListBoxUDB.Destroy;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  inherited;
end;

procedure TCustomListBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

procedure TCustomListBoxUDB.SetListConnection(Value: TCustomConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TCustomListBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TCustomListBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TCustomListBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TCustomListBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
end;

procedure TCustomListBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TCustomListBoxUDB.ReloadItems;
var
  id: integer;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=FListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      Items.BeginUpdate;
      try
        Items.Clear;
        id:=0;
        while not FListDataSet.Eof do
        begin
          if FListDataSet.FieldCount > 1 then
            id:=FListDataSet.Fields[1].AsInteger;
          Items.AddObject(FListDataSet.Fields[0].AsString, TObject(id));
          FListDataSet.Next;
        end;
      finally
        Items.EndUpdate;
      end;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TCustomListBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TCustomListBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

function TCustomListBoxUDB.GetItemObjectAsInt: integer;
begin
  if (ItemIndex >= 0) and Assigned(Items.Objects[ItemIndex]) then
    Result:=integer(Items.Objects[ItemIndex])
  else
    Result:=0;
end;

function TCustomListBoxUDB.SetActiveItem(ObjectAsInt: integer): integer;
begin
  ItemIndex:=Items.IndexOfObject(TObject(ObjectAsInt));
  Result:=ItemIndex;
end;

// ==============================================================================
// TrDBListBox

constructor TCustomrDBListBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);

  FListButton:=TBoundComboButton.CreateBtn(self, '', '', resnameDBLOOKUPSELECT, 1);
  FListButton.FreeNotification(self);
  FListButton.OnClick:=ComboButtonClick;

  FLinkedDSEvent:=TList.Create;

  FListAutoLoad:=true;
  FListButtonVisible:=true;
  FListButtonAllwaysEnabled:=true;
end;

destructor TCustomrDBListBoxUDB.Destroy;
var
  i: integer;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  FreeAndNil(FListButton);
  for i:=0 to FLinkedDSEvent.Count - 1 do
    try
      TCustomQueryUDB(FLinkedDSEvent[i]).RemoveEventNotify(DSEventCallBack);
    except
    end;
  FreeAndNil(FLinkedDSEvent);
  inherited;
end;

procedure TCustomrDBListBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TCustomrDBListBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
  if (AComponent = FListButton) and (Operation = opRemove) then
    FListButton:=nil;
end;

procedure TCustomrDBListBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TCustomrDBListBoxUDB.SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FListButton) then
    FListButton.RefreshPosition;
end;

procedure TCustomrDBListBoxUDB.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FListButton) then
    FListButton.Parent:=AParent;
end;

procedure TCustomrDBListBoxUDB.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled;
end;

procedure TCustomrDBListBoxUDB.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TCustomrDBListBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

function TCustomrDBListBoxUDB.GetListConnection: TCustomConnectionUDB;
begin
  if Assigned(FListConnection) then
    Result:=FListConnection
  else
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
      Result:=TCustomConnectionUDB(TCustomQueryUDB(DataSource.DataSet).Connection)
    else
      Result:=nil;
end;

procedure TCustomrDBListBoxUDB.SetListConnection(Value: TCustomConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TCustomrDBListBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TCustomrDBListBoxUDB.SetListButtonVisible(Value: boolean);
begin
  FListButtonVisible:=Value;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TCustomrDBListBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TCustomrDBListBoxUDB.ComboButtonClick(Sender: TObject);
var
  F: TField;
begin
  F:=nil;
  if FListDataSet.Active then
    F:=FListDataSet.Fields[0];
  if Assigned(FClickComboButton) then
    FClickComboButton(self, GetListConnection, ListButtonTableName, self.Field, F);
end;

procedure TCustomrDBListBoxUDB.DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
begin
  if EventType in FListAutoReloadEvents then
    if Assigned(DataSource) and (DataSet = DataSource.DataSet) then
      ReloadItems;
end;

procedure TCustomrDBListBoxUDB.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  if (FListAutoReloadEvents <> []) and Assigned(DataSource) and Assigned(DataSource.DataSet) and (DataSource.DataSet is TCustomQueryUDB) then
    if FLinkedDSEvent.IndexOf(DataSource.DataSet) < 0 then
    begin
      FLinkedDSEvent.Add(DataSource.DataSet);
      TCustomQueryUDB(DataSource.DataSet).AddEventNotify(DSEventCallBack);
    end;
  ro:=ReadOnly or (Assigned(Field) and Field.ReadOnly);
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled and not ro and (FListButtonAllwaysEnabled or (FDataLink.Active and FDataLink.Editing));
end;

procedure TCustomrDBListBoxUDB.ReloadItems;
var
  id, MaxLength: integer;
  s: string;
  F: TField;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=GetListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      Items.BeginUpdate;
      try
        Items.Clear;
        id:=0;
        while not FListDataSet.Eof do
        begin
          if FListDataSet.FieldCount > 1 then
            id:=FListDataSet.Fields[1].AsInteger;
          s:=FListDataSet.Fields[0].AsString;

          // limit length of text according field size
          MaxLength:=0;
          if Assigned(DataSource) and Assigned(DataSource.DataSet) then
          begin
            F:=DataSource.DataSet.FindField(DataField);
            if Assigned(F) and (F.DataType in [ftString, ftWideString]) and (MaxLength = 0) then
              MaxLength:=F.Size;
          end;
          if MaxLength > 0 then
            s:=copy(s, 1, MaxLength);

          Items.AddObject(FListDataSet.Fields[0].AsString, TObject(id)); FListDataSet.Next;
        end;
      finally
        Items.EndUpdate;
      end;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
    if Assigned(Field) and Field.DataSet.Active then
      try
        ItemIndex:=Items.IndexOf(Field.AsString);
      except
      end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TCustomrDBListBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TCustomrDBListBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

// ==============================================================================
// TrDBLookupListBox

constructor TCustomrDBLookupListBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListDataSource:=TDataSource.Create(self);

  FListButton:=TBoundComboButton.CreateBtn(self, '', '', resnameDBLOOKUPSELECT, 1);
  FListButton.FreeNotification(self);
  FListButton.OnClick:=ComboButtonClick;

  FLinkedDSEvent:=TList.Create;

  FListAutoLoad:=true;
  FListButtonVisible:=true;
  FListButtonAllwaysEnabled:=true;
end;

destructor TCustomrDBLookupListBoxUDB.Destroy;
var
  i: integer;
begin
  FListDataSet.Close;
  FListDataSource.DataSet:=nil;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  FreeAndNil(FListDataSource);
  FreeAndNil(FListButton);
  for i:=0 to FLinkedDSEvent.Count - 1 do
    try
      TCustomQueryUDB(FLinkedDSEvent[i]).RemoveEventNotify(DSEventCallBack);
    except
    end;
  FreeAndNil(FLinkedDSEvent);
  inherited;
end;

procedure TCustomrDBLookupListBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TCustomrDBLookupListBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
  if (AComponent = FListButton) and (Operation = opRemove) then
    FListButton:=nil;
end;

procedure TCustomrDBLookupListBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TCustomrDBLookupListBoxUDB.SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FListButton) then
    FListButton.RefreshPosition;
end;

procedure TCustomrDBLookupListBoxUDB.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FListButton) then
    FListButton.Parent:=AParent;
end;

procedure TCustomrDBLookupListBoxUDB.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled;
end;

procedure TCustomrDBLookupListBoxUDB.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TCustomrDBLookupListBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

function TCustomrDBLookupListBoxUDB.GetListConnection: TCustomConnectionUDB;
begin
  if Assigned(FListConnection) then
    Result:=FListConnection
  else
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
      Result:=TCustomConnectionUDB(TCustomQueryUDB(DataSource.DataSet).Connection)
    else
      Result:=nil;
end;

procedure TCustomrDBLookupListBoxUDB.SetListConnection(Value: TCustomConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TCustomrDBLookupListBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TCustomrDBLookupListBoxUDB.SetListButtonVisible(Value: boolean);
begin
  FListButtonVisible:=Value;
  if Assigned(FListButton) then
    FListButton.Visible:=Visible and FListButtonVisible;
end;

procedure TCustomrDBLookupListBoxUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TCustomrDBLookupListBoxUDB.ComboButtonClick(Sender: TObject);
var
  F: TField;
begin
  F:=nil;
  if FListDataSet.Active then
    F:=FListDataSet.Fields[0];
  if Assigned(FClickComboButton) then
    FClickComboButton(self, GetListConnection, ListButtonTableName, Field, F);
end;

procedure TCustomrDBLookupListBoxUDB.DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
begin
  if EventType in FListAutoReloadEvents then
    if Assigned(DataSource) and (DataSet = DataSource.DataSet) then
      ReloadItems;
end;

procedure TCustomrDBLookupListBoxUDB.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  if (FListAutoReloadEvents <> []) and Assigned(DataSource) and Assigned(DataSource.DataSet) and (DataSource.DataSet is TCustomQueryUDB) then
    if FLinkedDSEvent.IndexOf(DataSource.DataSet) < 0 then
    begin
      FLinkedDSEvent.Add(DataSource.DataSet);
      TCustomQueryUDB(DataSource.DataSet).AddEventNotify(DSEventCallBack);
    end;
  ro:=ReadOnly or (Assigned(Field) and Field.ReadOnly);
  if Assigned(FListButton) then
    FListButton.Enabled:=Enabled and not ro and (FListButtonAllwaysEnabled or (FDataLink.Active and FDataLink.Editing));
end;

procedure TCustomrDBLookupListBoxUDB.ReloadItems;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=GetListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    FListDataSet.LoadFieldsProps;
    if not(csDesigning in ComponentState) then
    begin
      FListDataSource.DataSet:=FListDataSet;
      ListSource:=FListDataSource;
      KeyField:=FListDataSet.Fields[0].FieldName;
      ListField:=FListDataSet.Fields[1].FieldName;
      DBEditLabel.SetLabelField(ListSource, ListField);
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TCustomrDBLookupListBoxUDB.GetListValue: string;
var
  F: TField;
begin
  F:=FListDataSet.FindField(ListField);
  if Assigned(F) and not VarIsNull(KeyValue) then
    Result:=F.AsString
  else
    Result:='';
end;

function TCustomrDBLookupListBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TCustomrDBLookupListBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

// ==============================================================================
// TCustomCheckListBoxUDB

constructor TCustomCheckListBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSet.SetReadOnly(true);
  FListAutoLoad:=true;
end;

destructor TCustomCheckListBoxUDB.Destroy;
begin
  FListDataSet.Close;
  FreeAndNil(FListSQL);
  FreeAndNil(FListDataSet);
  inherited;
end;

procedure TCustomCheckListBoxUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

procedure TCustomCheckListBoxUDB.SetListConnection(Value: TCustomConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TCustomCheckListBoxUDB.SetListSQLTryNow(Value: boolean);
begin
  FListSQLTryNow:=Value;
  try
    if FListSQLTryNow then
      ReloadItems;
  finally
    FListSQLTryNow:=false;
  end;
end;

procedure TCustomCheckListBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadItems;
end;

procedure TCustomCheckListBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
end;

procedure TCustomCheckListBoxUDB.DoEnter;
begin
  if FListAutoLoad and not FListDataSet.Active then
    ReloadItems;
  inherited;
end;

procedure TCustomCheckListBoxUDB.ReloadItems;
var
  id: integer;
begin
  try
    FListDataSet.Close;
    FListDataSet.Connection:=FListConnection;
    if Assigned(FGetListSQL) then
      FGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    if not(csDesigning in ComponentState) then
    begin
      Items.BeginUpdate;
      try
        Items.Clear;
        id:=0;
        while not FListDataSet.Eof do
        begin
          if FListDataSet.FieldCount > 1 then
            id:=FListDataSet.Fields[1].AsInteger;
          Items.AddObject(FListDataSet.Fields[0].AsString, TObject(id));
          FListDataSet.Next;
        end;
      finally
        Items.EndUpdate;
      end;
    end
    else
    begin
      FListDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

function TCustomCheckListBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TCustomCheckListBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FListDataSet) or not FListDataSet.Active then
    ReloadItems;
  if not FListDataSet.Active then
    Exit;
  if FieldIndex < FListDataSet.FieldCount then
    Result:=FListDataSet.Fields[FieldIndex];
end;

function TCustomCheckListBoxUDB.GetItemObjectAsInt: integer;
begin
  if (ItemIndex >= 0) and Assigned(Items.Objects[ItemIndex]) then
    Result:=integer(Items.Objects[ItemIndex])
  else
    Result:=0;
end;

function TCustomCheckListBoxUDB.SetActiveItem(ObjectAsInt: integer): integer;
begin
  ItemIndex:=Items.IndexOfObject(TObject(ObjectAsInt));
  Result:=ItemIndex;
end;

// ==============================================================================
// TCustomrDBCheckListBoxUDB

constructor TCustomrDBCheckListBoxUDB.Create(AOwner: TComponent);
begin
  inherited;
  FItemsListDataSet:=TQueryIntList.Create(self);
  FItemsListSQL:=TStringList.Create;
  FItemsListAutoLoad:=true;
  FItemsValueDataSet:=TQueryIntList.Create(self);
  FItemsValueSQL:=TStringList.Create;
  FLinkedDSEvent:=TList.Create;
end;

destructor TCustomrDBCheckListBoxUDB.Destroy;
var
  i: integer;
begin
  FItemsListDataSet.Close;
  FreeAndNil(FItemsListDataSet);
  FItemsValueDataSet.Close;
  FreeAndNil(FItemsValueDataSet);
  FreeAndNil(FItemsListSQL);
  FreeAndNil(FItemsValueSQL);
  for i:=0 to FLinkedDSEvent.Count - 1 do
    try
      TCustomQueryUDB(FLinkedDSEvent[i]).RemoveEventNotify(DSEventCallBack);
    except
    end;
  FreeAndNil(FLinkedDSEvent);
  inherited;
end;

procedure TCustomrDBCheckListBoxUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FItemsListAutoLoad then
      ReloadItemsList(false);
end;

procedure TCustomrDBCheckListBoxUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
end;

function TCustomrDBCheckListBoxUDB.GetListConnection: TCustomConnectionUDB;
begin
  if Assigned(FListConnection) then
    Result:=FListConnection
  else
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
      Result:=TCustomConnectionUDB(TCustomQueryUDB(DataSource.DataSet).Connection)
    else
      Result:=nil;
end;

procedure TCustomrDBCheckListBoxUDB.SetListConnection(Value: TCustomConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TCustomrDBCheckListBoxUDB.SetItemsListSQL(Value: TStrings);
begin
  FItemsListSQL.Assign(Value);
end;

procedure TCustomrDBCheckListBoxUDB.SetItemsListSQLTryNow(Value: boolean);
begin
  FItemsListSQLTryNow:=Value;
  try
    if FItemsListSQLTryNow then
      ReloadItemsList(false);
  finally
    FItemsListSQLTryNow:=false;
  end;
end;

procedure TCustomrDBCheckListBoxUDB.SetItemsValueSQL(Value: TStrings);
begin
  FItemsValueSQL.Assign(Value);
end;

procedure TCustomrDBCheckListBoxUDB.SetItemsValueSQLTryNow(Value: boolean);
begin
  FItemsValueSQLTryNow:=Value;
  try
    if FItemsValueSQLTryNow then
      ReloadItemsValue;
  finally
    FItemsValueSQLTryNow:=false;
  end;
end;

procedure TCustomrDBCheckListBoxUDB.SetItemsValueDataSetEvent(Value: TDataSetEvent);
begin
  FItemsValueDataSetEvent:=Value;
  if Assigned(FItemsValueDataSet) then
    FItemsValueDataSet.OnDataSetEvent:=Value;
end;

procedure TCustomrDBCheckListBoxUDB.DSEventCallBack(DataSet: TDataSet; EventType: TDatasetEventType);
begin
  if Assigned(DataSource) and (DataSet = DataSource.DataSet) then
    case EventType of
      deAfterPostOrApplyUpdates:
        if PostponedSaving then
          DoSaveItemValues;
      deAfterCancel:
        DoCancelItemValues;
    end;
end;

procedure TCustomrDBCheckListBoxUDB.ReloadItemsList(ReloadValues: boolean);
begin
  try
    FItemsListDataSet.Close;
    FItemsListDataSet.Connection:=GetListConnection;
    if Assigned(FGetItemsListSQL) then
      FGetItemsListSQL(self, ItemsListSQL);
    FItemsListDataSet.SQL.Assign(ItemsListSQL);
    FItemsListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    // FItemsListDataSet.LoadFieldsProps;
    if not(csDesigning in ComponentState) then
    begin
      FAllItems.Clear;
      while not FItemsListDataSet.Eof do
      begin
        FAllItems.AddCheckItem(FItemsListDataSet.Fields[0].AsInteger, FItemsListDataSet.Fields[1].AsString, false);
        FItemsListDataSet.Next;
      end;
    end
    else
    begin
      FItemsListDataSet.Close;
      Beep;
    end;
    if ReloadValues then
      DoLoadItemValues;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

procedure TCustomrDBCheckListBoxUDB.ReloadItemsValue;
begin
  try
    FItemsValueDataSet.Close;
    FItemsValueDataSet.Connection:=GetListConnection;
    // add notify link
    if Assigned(DataSource) and Assigned(DataSource.DataSet) and (DataSource.DataSet is TCustomQueryUDB) then
      if FLinkedDSEvent.IndexOf(DataSource.DataSet) < 0 then
      begin
        FLinkedDSEvent.Add(DataSource.DataSet);
        TCustomQueryUDB(DataSource.DataSet).AddEventNotify(DSEventCallBack);
      end;

    if Assigned(FDataLink.Field) then
    begin
      // add where condition if not used
      if rdbtool.GetSQLWhere(FItemsValueSQL) = '' then
        FItemsValueSQL.Add('WHERE 1=0');
      // set where cond according master field
      rdbtool.SetSQLWhere(FItemsValueSQL, DataField + '=' + IntToStr(FDataLink.Field.AsInteger));
    end;
    if Assigned(FGetItemsValueSQL) then
      FGetItemsValueSQL(self, FItemsValueSQL);
    FItemsValueDataSet.SQL.Assign(FItemsValueSQL);
    FItemsValueDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    // FItemsValueDataSet.LoadFieldsProps;
    if not(csDesigning in ComponentState) then
    begin
      FAllItems.ResetItemsChecked;
      FItemsValueDataSet.First;
      while not FItemsValueDataSet.Eof do
      begin
        FAllItems.SetItemChecked(FItemsValueDataSet.Fields[1].AsInteger, true);
        FItemsValueDataSet.Next;
      end;
      FilterItemsList;
    end
    else
    begin
      FItemsValueDataSet.Close;
      Beep;
    end;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

procedure TCustomrDBCheckListBoxUDB.DoLoadItemValues;
begin
  if not PostponedSaving then
    ReloadItemsValue;
  inherited;
end;

procedure TCustomrDBCheckListBoxUDB.DoSaveItemValues;
var
  i, id: integer;
begin
  if Assigned(FOnBeforeSaveItemValues) then
    FOnBeforeSaveItemValues(self, FAllItems);
  for i:=0 to FAllItems.Count - 1 do
  begin
    id:=FAllItems.CheckItem(i).id;
    if FAllItems.CheckItem(i).Checked then
    begin
      if not LocateRecord(FItemsValueDataSet.Fields[1], id) then
      begin
        FItemsValueDataSet.Append;
        FItemsValueDataSet.Fields[0].AsInteger:=FDataLink.Field.AsInteger;
        FItemsValueDataSet.Fields[1].AsInteger:=id;
        FItemsValueDataSet.Post;
      end;
    end
    else
    begin
      if LocateRecord(FItemsValueDataSet.Fields[1], id) then
        FItemsValueDataSet.Delete;
    end;
  end;
  inherited;
end;

function TCustomrDBCheckListBoxUDB.GetListField(const FieldName: string): TField;
begin
  if Assigned(GetListField(0)) then
    Result:=FItemsListDataSet.FieldByName(FieldName)
  else
    Result:=nil;
end;

function TCustomrDBCheckListBoxUDB.GetListField(FieldIndex: integer): TField;
begin
  Result:=nil;
  if not Assigned(FItemsListDataSet) or not FItemsListDataSet.Active then
    DoLoadItemList;
  if not FItemsListDataSet.Active then
    Exit;
  if FieldIndex < FItemsListDataSet.FieldCount then
    Result:=FItemsListDataSet.Fields[FieldIndex];
end;

function TCustomrDBCheckListBoxUDB.GetItemObjectAsInt: integer;
begin
  if (ItemIndex >= 0) and Assigned(Items.Objects[ItemIndex]) then
    Result:=integer(Items.Objects[ItemIndex])
  else
    Result:=0;
end;

function TCustomrDBCheckListBoxUDB.SetActiveItem(ObjectAsInt: integer): integer;
begin
  ItemIndex:=Items.IndexOfObject(TObject(ObjectAsInt));
  Result:=ItemIndex;
end;

// ==============================================================================
// TCustomrDBRecViewUDB

constructor TCustomrDBRecViewUDB.Create(AOwner: TComponent);
begin
  inherited;
  FRecSQL:=TStringList.Create;
  FRecDataSet:=TQueryIntList.Create(self);
  FRecDataSource:=TDataSource.Create(self);
  FRecAutoLoad:=true;
end;

destructor TCustomrDBRecViewUDB.Destroy;
begin
  FRecDataSource.DataSet:=nil;
  FDataLink.DataSource:=nil;
  FRecDataSet.Close;
  FreeAndNil(FRecDataSource);
  FreeAndNil(FRecDataSet);
  FreeAndNil(FRecSQL);
  inherited;
end;

procedure TCustomrDBRecViewUDB.Loaded;
begin
  inherited;
  FTemplateDataSource:=DataSource;
  if not(csDesigning in ComponentState) then
    if FRecAutoLoad then
      ReloadData;
end;

procedure TCustomrDBRecViewUDB.SetRecSQL(Value: TStrings);
begin
  FRecSQL.Assign(Value);
end;

procedure TCustomrDBRecViewUDB.SetRecConnection(Value: TCustomConnectionUDB);
begin
  FRecConnection:=Value;
  if Assigned(FRecConnection) then
    FRecConnection.FreeNotification(self);
end;

procedure TCustomrDBRecViewUDB.SetRecSQLTryNow(Value: boolean);
begin
  // reset RecSQLTryNow during loading
  if (csLoading in ComponentState) then
    Value:=false;
  FRecSQLTryNow:=Value;
  if FRecSQLTryNow then
    ReloadData
  else
    FRecDataSet.Close;
end;

procedure TCustomrDBRecViewUDB.SetRecDataSetEvent(Value: TDataSetEvent);
begin
  FRecDataSetEvent:=Value;
  if Assigned(FRecDataSet) then
    FRecDataSet.OnDataSetEvent:=Value;
end;

procedure TCustomrDBRecViewUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FRecConnection) and (Operation = opRemove) then
    FRecConnection:=nil;
end;

procedure TCustomrDBRecViewUDB.ReloadData;
var
  i: integer;
  F, T: TField;
  L: TStringList;
begin
  if FRecSQL.Count = 0 then
    Exit;
  try
    FRecDataSet.Close;
    FRecDataSet.SetReadOnly(FRecDataSetReadOnly);
    FRecDataSet.Connection:=FRecConnection;
    if Assigned(FOnGetRecSQL) then
      FOnGetRecSQL(self, RecSQL);
    FRecDataSet.SQL.Assign(FRecSQL);
    FRecDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    // keep open in design time to allow set properties in DT dialog
    if Assigned(FTemplateDataSource) and Assigned(FTemplateDataSource.DataSet) then
    begin
      L:=TStringList.Create;
      try
        // copy fields props
        for i:=0 to FTemplateDataSource.DataSet.FieldCount - 1 do
        begin
          F:=FTemplateDataSource.DataSet.Fields[i];
          L.Add(F.FieldName + '=' + F.DisplayLabel);
          T:=FRecDataSet.FindField(F.FieldName);
          if Assigned(T) then
          begin
            T.Visible:=F.Visible;
            T.DisplayWidth:=F.DisplayWidth;
            if T is TNumericField then
              TNumericField(T).DisplayFormat:=TNumericField(F).DisplayFormat;
            if T is TBooleanField then
              TBooleanField(T).DisplayValues:=TBooleanField(F).DisplayValues;
          end;
        end;
        Fields.Assign(L);
      finally
        FreeAndNil(L);
      end;
    end
    else
    begin
      FRecDataSet.SetFieldDefaultFormat;
      FRecDataSet.LoadFieldsProps;
    end;
    FRecDataSource.DataSet:=FRecDataSet;
    FDataLink.DataSource:=FRecDataSource;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

// ==============================================================================
// TCustomrDBGrid_MSUDB

constructor TCustomrDBGrid_MSUDB.Create(AOwner: TComponent);
begin
  inherited;
  FListSQL:=TStringList.Create;
  FListDataSet:=TQueryIntList.Create(self);
  FListDataSource:=TDataSource.Create(self);
  FListAutoLoad:=true;
end;

destructor TCustomrDBGrid_MSUDB.Destroy;
begin
  FListDataSource.DataSet:=nil;
  FListDataSet.Close;
  FreeAndNil(FListDataSource);
  FreeAndNil(FListDataSet);
  FreeAndNil(FListSQL);
  inherited;
end;

procedure TCustomrDBGrid_MSUDB.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
    if FListAutoLoad then
      ReloadData;
end;

procedure TCustomrDBGrid_MSUDB.SetListSQL(Value: TStrings);
begin
  FListSQL.Assign(Value);
end;

procedure TCustomrDBGrid_MSUDB.SetListConnection(Value: TCustomConnectionUDB);
begin
  FListConnection:=Value;
  if Assigned(FListConnection) then
    FListConnection.FreeNotification(self);
end;

procedure TCustomrDBGrid_MSUDB.SetListSQLTryNow(Value: boolean);
begin
  // reset ListSQLTryNow during loading
  if (csLoading in ComponentState) then
    Value:=false;
  FListSQLTryNow:=Value;
  if FListSQLTryNow then
    ReloadData
  else
    FListDataSet.Close;
end;

procedure TCustomrDBGrid_MSUDB.SetListDataSetEvent(Value: TDataSetEvent);
begin
  FListDataSetEvent:=Value;
  if Assigned(FListDataSet) then
    FListDataSet.OnDataSetEvent:=Value;
end;

procedure TCustomrDBGrid_MSUDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FListConnection) and (Operation = opRemove) then
    FListConnection:=nil;
end;

procedure TCustomrDBGrid_MSUDB.ReloadData;
begin
  if FListSQL.Count = 0 then
    Exit;
  try
    FListDataSet.Close;
    FListDataSet.SetReadOnly(FListDataSetReadOnly);
    FListDataSet.Connection:=FListConnection;
    if Assigned(FOnGetListSQL) then
      FOnGetListSQL(self, ListSQL);
    FListDataSet.SQL.Assign(FListSQL);
    FListDataSet.RequeryEx(RosiCompConfig.rDBComponentsCfg.ReloadItemsShowCursor);
    // keep open in design time to allow set properties in DT dialog
    FListDataSet.SetFieldDefaultFormat;
    FListDataSet.LoadFieldsProps;
    FListDataSource.DataSet:=FListDataSet;
    DataSource:=FListDataSource;
  except
    on E: Exception do
      MessageDlg(self.Name + #13 + E.Message, mtError, [mbOk], 0);
  end;
end;

// ==============================================================================
// TrDBGridUDBSorter

function TCustomrDBGridSorter_UDB.SetSort_CreateIndex(DataSet: TDataSet; const SortString: string; const SortFields: string; const DescendingFields: string)
  : boolean;
var
{$IFDEF UDB_ADO} DS: TCustomADODataSet; {$ENDIF}
{$IFDEF UDB_DBX} DS: TSimpleDataSet;
  idxName: string; {$ENDIF}
{$IFDEF UDB_IB} DS: TIBCustomDataSet; {$ENDIF}
{$IFDEF UDB_FIB} DS: TpFIBDataSet {$ENDIF}
{$IFDEF UDB_Zeos} DS: TZAbstractDataSet; {$ENDIF}
{$IFDEF UDB_FD} DS: TFDDataSet;
  idxName: string; {$ENDIF}
begin
{$IFDEF UDB_ADO}
  if (DataSet is TCustomADODataSet) then
  begin
    DS:=DataSet as TCustomADODataSet;
    DS.Sort:=SortString;
    Result:=true;
  end
  else
    Result:= inherited;
{$ENDIF}
{$IFDEF UDB_Zeos}
  if (DataSet is TZAbstractDataSet) then
  begin
    DS:=DataSet as TZAbstractDataSet;
    DS.SortedFields:=SortString;
    Result:=true;
  end
  else
    Result:= inherited;
{$ENDIF}
{$IFDEF UDB_DBX}
  if (DataSet is TSimpleDataSet) then
  begin
    DS:=DataSet as TSimpleDataSet;
    // indexname can be max 31 chars long - use hash from fields names
    idxName:=GetIndexNameHash(SortFields + '_' + DescendingFields);
    if DS.IndexDefs.IndexOf(idxName) < 0 then
      DS.AddIndex(idxName, SortFields, [], DescendingFields);
    DS.IndexDefs.Update;
    DS.IndexName:=idxName;
    Result:=true;
  end
  else
    Result:= inherited;
{$ENDIF}
{$IFDEF UDB_IB}
  Result:= inherited;
{$ENDIF}
{$IFDEF UDB_FIB}
  Result:= inherited;
{$ENDIF}
{$IFDEF UDB_FD}
  if (DataSet is TFDDataSet) then
  begin
    DS:=DataSet as TFDDataSet;
    // indexname can be max 31 chars long - use hash from fields names
    idxName:=SortFields + '_' + DescendingFields;
    if DS.IndexDefs.IndexOf(idxName) < 0 then
      DS.AddIndex(idxName, SortFields, '', [], DescendingFields);
    DS.IndexDefs.Update;
    DS.IndexName:=idxName;
    Result:=true;
  end
  else
    Result:= inherited;
{$ENDIF}
end;

procedure TCustomrDBGridSorter_UDB.DatasetRequery(DataSet: TDataSet);
begin
{$IFDEF UDB_ADO}
  if (DataSet is TCustomADODataSet) then
    TCustomADODataSet(DataSet).Requery()
  else
{$ENDIF}
    inherited;
end;

end.
